# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Image processing parameters.

The following symbols are imported in the equimage/equimagelab namespaces for convenience:
  "get_RGB_luma", "set_RGB_luma".
"""

__all__ = ["get_RGB_luma", "set_RGB_luma"]

import numpy as np

# Use imageio or imread module ?

IMAGEIO = False

# Image type.

imagetype = np.float32

def get_image_type():
  """Return the image type.

  Returns:
    str: The image type, either "float32" (for 32 bits floats) or "float64" (for 64 bits floats).
  """
  return "float32" if imagetype == np.float32 else "float64"

def set_image_type(dtype):
  """Set image type.

  Args:
    dtype (str): The image type.
      Can be either "float32" (for 32 bits floats) or "float64" (for 64 bits floats).
  """
  global imagetype
  if dtype == "float32":
    imagetype = np.float32
  elif dtype == "float64":
    imagetype = np.float64
  else:
    raise ValueError(f"Error, the image type must be 'float32' or 'float64' (got '{dtype}').")

# CIE color spaces parameters.

CIEilluminant = "D65" # Illuminant.
CIEobserver = "2" # Observer.

def get_CIE_params():
  """Return the CIE illuminant and observer.

  Returns:
    str, str: The CIE illuminant and observer as a tuple of strings.
  """
  return CIEilluminant, CIEobserver

def set_CIE_params(illuminant, observer):
  """Set CIE illuminant and observer.

  Args:
    illuminant (str): The name of the standard illuminant.
      Can be "A", "B", "C", "D50", "D55", "D65", "D75", or "E".
      See https://en.wikipedia.org/wiki/Standard_illuminant.
    observer (str): The name of the observer.
      Can be "2" (2-degree observer) or "10" (10-degree observer).
  """
  if illuminant not in ["A", "B", "C", "D50", "D55", "D65", "D75", "E"]:
    raise ValueError("Error, the CIE illuminant must be 'A', 'B', 'C', 'D50', 'D55', 'D65', 'D75' or 'E'.")
  if observer not in ["2", "10"]:
    raise ValueError("Error, the observer must be '2' or '10'.")
  global CIEilluminant, CIEobserver
  CIEilluminant = illuminant
  CIEobserver = observer

# Weights of the RGB components in the luma.

rgbluma = (.212671, .715160, .072169)

def get_RGB_luma():
  """Return the RGB weights rgbluma of the luma.

  The luma L of an image is the average of the RGB components weighted by rgbluma:

    L = rgbluma[0]*image[0]+rgbluma[1]*image[1]+rgbluma[2]*image[2]

  Returns:
    float, float, float: The red, blue, green weights rgbluma of the luma.
  """
  return rgbluma

def set_RGB_luma(rgb, verbose = True):
  """Set the RGB weights of the luma.

  Args:
    rgb: The RGB weights of the luma as:

      - a tuple, list or array of the (red, green, blue) weights. They will be normalized so that
        their sum is 1.
      - the string "uniform": the RGB weights are set to (1/3, 1/3, 1/3).
      - the string "human": the RGB weights are set to (.212671, .715160, .072169). The luma is then
        the luminance for lRGB images, and an approximate substitute for the lightness for sRGB images.

    verbose (bool, optional): If True (default), print the updated definition of
      the luma.
  """
  if isinstance(rgb, str):
    if rgb == "uniform":
      set_RGB_luma((1./3., 1./3., 1./3.))
    elif rgb == "human":
      set_RGB_luma((.212671, .715160, .072169))
    else:
      raise ValueError("Error, the input rgb weights must be a tuple/list/array with three scalar elements, the string 'uniform' or the string 'human'.")
  else:
    w = np.array(rgb)
    if w.shape != (3,): raise ValueError("Error, the input rgb weights must be a tuple/list/array with three scalar elements, the string 'uniform' or the string 'human'.")
    if any(w < 0.): raise ValueError("Error, the input rgb weights must be >= 0.")
    s = np.sum(w)
    if s == 0.: raise ValueError("Error, the sum of the input rgb weights must be > 0.")
    global rgbluma ; rgbluma = w/s
    if verbose: print(f"Luma = {rgbluma[0]:.4f}R+{rgbluma[1]:.4f}G+{rgbluma[2]:.4f}B.")

# Number of bins in the histograms.

maxhistbins = 8192 # Maximum number of bins within [0, 1].
defhistbins = 0    # Default number of bins within [0, 1] (see set_default_hist_bins).

def set_max_hist_bins(n):
  """Set the maximum number of bins in the histograms.

  Args:
    n (int): The maximum number of bins within [0, 1].
  """
  global maxhistbins
  maxhistbins = n

def set_default_hist_bins(n):
  """Set the default number of bins in the histograms.

  Args:
    n (int): If strictly positive, the default number of bins within [0, 1] (practically
      limited to `equimage.params.maxhistbins`). If zero, the number of bins is computed
      according to the statistics of each image. If strictly negative, the number of bins
      is set to `equimage.params.maxhistbins`.
  """
  global defhistbins
  defhistbins = n if n >= 0 else maxhistbins

# Number of x mesh points for transformation function plots y = f(x in [0, 1]).

ntranslo = 128 # Low  resolution.
ntranshi = 256 # High resolution.

# Exclude pixels <= 0 or >= 1 from the percentiles in image statistics ?

exclude01 = False

# Default wavelet.

defwavelet = "coif4"



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Image statistics & histograms."""

import numpy as np

from . import params
from . import helpers

def parse_channels(channels, errors = True):
  """Parse channel keys.

  Args:
    channels (str): A combination of channel keys:

      - "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
        The first/second/third channel (all images).
      - "V": The HSV value (RGB, HSV and grayscale images).
      - "S": The HSV saturation (RGB, HSV and grayscale images).
      - "L'": The HSL lightness (RGB, HSL and grayscale images).
      - "S'": The HSL saturation (RGB, HSL and grayscale images).
      - "H": The HSV hue (RGB, HSV and grayscale images).
      - "L": The luma (RGB and grayscale images).
      - "L*": The CIE lightness L* (RGB, grayscale, CIELab and CIELuv images).
      - "c*": The CIE chroma c* (CIELab and CIELuv images).
      - "s*": The CIE saturation s* (CIELuv images).
      - "h*": The CIE hue angle h* (CIELab and CIELuv images).

    errors (bool, optional): If False, discard unknown channel keys;
      If True (default), raise a ValueError.

  Returns:
    list: The list of channel keys.
  """
  keys = []
  prevkey = None
  for key in channels:
    ok = False
    if key in ["1", "2", "3", "R", "G", "B", "L", "V", "S", "H", "c", "s", "h"]:
      keys.append(key)
      ok = True
    elif key == "'":
      if prevkey in ["L", "S"]:
        keys[-1] += key
        ok = True
    elif key == "*":
      if prevkey in ["L", "c", "s", "h"]:
        keys[-1] += key
        ok = True
    elif key == " ":
      ok = True # Skip spaces.
    if not ok and errors: raise ValueError(f"Error, unknown channel '{key}'.")
    prevkey = key
  return keys

#####################################
# For inclusion in the Image class. #
#####################################

class MixinImage:
  """To be included in the Image class."""

  def histograms(self, channels = None, nbins = None, recompute = False):
    """Compute histograms of selected channels of the image.

    The histograms are both returned and embedded in the object as self.hists. Histograms already
    registered in self.hists are not recomputed unless required.

    See also:
      :meth:`params.set_default_hist_bins() <equimage.params.set_default_hist_bins>`,
      :meth:`params.set_max_hist_bins() <equimage.params.set_max_hist_bins>`

    Args:
      channels (str, optional): A combination of keys for the selected channels:

        - "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          The first/second/third channel (all images).
        - "V": The HSV value (RGB, HSV and grayscale images).
        - "S": The HSV saturation (RGB, HSV and grayscale images).
        - "L'": The HSL lightness (RGB, HSL and grayscale images).
        - "S'": The HSL saturation (RGB, HSL and grayscale images).
        - "H": The HSV hue (RGB, HSV and grayscale images).
        - "L": The luma (RGB and grayscale images).
        - "L*": The CIE lightness L* (RGB, grayscale, CIELab and CIELuv images).
        - "c*": The CIE chroma c* (CIELab and CIELuv images).
        - "s*": The CIE saturation s* (CIELuv images).
        - "h*": The CIE hue angle h* (CIELab and CIELuv images).

        If it ends with a "+", channels gets appended with the keys already registered in self.hists.
        Default (if None) is "RGBL" for RGB images, "VS" for HSV images, "L'S'" for HSL images,
        "L" for grayscale images, "L*c*" for CIELab and "L*s*" for CIELuv images.
      nbins (int, optional): Number of bins within [0, 1] in the histograms.
        Set to `equimage.params.maxhistbins` if negative, and computed from the image statistics
        using Scott's rule if zero. If None (default), set to `equimage.params.defhistbins`.
      recompute (bool, optional): If False (default), the histograms already registered in self.hists
        are not recomputed (provided they match nbins). If True, all histograms are recomputed.

    Returns:
      dict: hists[key] for key in channels, with:

        - hists[key].name = channel name (provided for convenience).
        - hists[key].nbins = number of bins within [0, 1].
        - hists[key].edges = histogram bins edges.
        - hists[key].counts = histogram bins counts.
        - hists[key].color = suggested line color for histogram plots.
    """
    if channels is None:
      if self.colorspace == "CIELab":
        channels = "L*c*"
      elif self.colorspace == "CIELuv":
        channels = "L*s*"
      elif self.colorspace == "lRGB" or self.colorspace == "sRGB":
        if self.colormodel == "RGB":
          channels = "RGBL"
        elif self.colormodel == "HSV":
          channels = "VS"
        elif self.colormodel == "HSL":
          channels = "V'S'"
        elif self.colormodel == "gray":
          channels = "L"
        else:
          self.color_model_error()
      else:
        self.color_space_error()
    if nbins is None: nbins = params.defhistbins
    if nbins == 0:
      if not recompute and hasattr(self, "hists"): # Retrieve the number of bins from the existing histograms.
        nbins = list(self.hists.values())[0].nbins
      else: # Compute the number of bins using Scott's rule.
        width, height = self.get_size()
        npixels = width*height
        if self.colormodel == "RGB":
          channel = (self.image[0]+self.image[1]+self.image[2])/3.
        elif self.colormodel == "HSV" or self.colormodel == "HSL":
          channel = self.image[2]
        else:
          channel = self.image[0]
        nbins = int(np.ceil(npixels**(1./3.)/(3.5*np.std(channel))))
    elif nbins < 0:
      nbins = params.maxhistbins
    nbins = min(max(nbins, 16), params.maxhistbins)
    if not hasattr(self, "hists"): self.hists = {} # Register empty histograms in the object, if none already computed.
    if channels and channels[-1] == "+":
      keys = parse_channels(channels[:-1])
      for key in self.hists.keys(): # Add missing keys.
        if not key in keys: keys.append(key)
    else:
      keys = parse_channels(channels)
    hists = {}
    for key in keys:
      if key in hists: # Already selected.
        print(f"Warning, channel '{key}' selected twice or more...")
        continue
      if not recompute and key in self.hists: # Already computed.
        if self.hists[key].nbins == nbins:
          hists[key] = self.hists[key]
          continue
      if key == "1":
        name = "Channel #1"
        color = "red"
        channel = self.image[0]
      elif key == "2":
        name = "Channel #2"
        color = "green"
        if self.get_nc() < 2: raise self.color_model_error()
        channel = self.image[1]
      elif key == "3":
        name = "Channel #3"
        color = "blue"
        if self.get_nc() < 3: raise self.color_model_error()
        channel = self.image[2]
      elif key == "R":
        self.check_color_model("RGB", "gray")
        name = "Red"
        color = "red"
        channel = self.image[0]
      elif key == "G":
        self.check_color_model("RGB", "gray")
        name = "Green"
        color = "green"
        channel = self.image[1] if self.colormodel == "RGB" else self.image[0]
      elif key == "B":
        self.check_color_model("RGB", "gray")
        name = "Blue"
        color = "blue"
        channel = self.image[2] if self.colormodel == "RGB" else self.image[0]
      elif key == "V":
        name = "HSV value"
        color = "darkslategray"
        channel = self.HSV_value()
      elif key == "S":
        name = "HSV saturation"
        color = "orange"
        channel = self.HSV_saturation()
      elif key == "L'":
        name = "HSL lightness"
        color = "darkslategray"
        channel = self.HSL_lightness()
      elif key == "S'":
        name = "HSL saturation"
        color = "orange"
        channel = self.HSL_saturation()
      elif key == "H":
        name = "HSX hue"
        color = "limegreen"
        channel = self.HSX_hue()
      elif key == "L":
        name = "Luma"
        color = "lightslategray"
        channel = self.luma()
      elif key == "L*":
        name = "Lightness"
        color = "lightsteelblue"
        channel = self.lightness()
      elif key == "c*":
        name = "CIE chroma"
        color = "orange"
        channel = self.CIE_chroma()
      elif key == "s*":
        name = "CIE saturation"
        color = "orange"
        channel = self.CIE_saturation()
      elif key == "h*":
        name = "CIE hue"
        color = "limegreen"
        channel = self.CIE_hue()
      else:
        raise ValueError(f"Error, unknown channel '{key}'.")
      mmin = np.floor(np.min(channel)*nbins)
      mmax = np.ceil (np.max(channel)*nbins)
      mbins = max(int(round(mmax-mmin)), 1)
      hists[key] = helpers.Container()
      hists[key].name = name
      hists[key].nbins = nbins
      hists[key].counts, hists[key].edges = np.histogram(channel, bins = mbins, range = (mmin/nbins, mmax/nbins), density = False)
      hists[key].color = color
    self.hists = hists
    return hists

  def statistics(self, channels = None, exclude01 = None, recompute = False):
    """Compute statistics of selected channels of the image.

    The statistics are both returned and embedded in the object as self.stats. Statistics already
    registered in self.stats are not recomputed unless required.

    Args:
      channels (str, optional): A combination of keys for the selected channels:

        - "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          The first/second/third channel (all images).
        - "V": The HSV value (RGB, HSV and grayscale images).
        - "S": The HSV saturation (RGB, HSV and grayscale images).
        - "L'": The HSL lightness (RGB, HSL and grayscale images).
        - "S'": The HSL saturation (RGB, HSL and grayscale images).
        - "H": The HSV hue (RGB, HSV and grayscale images).
        - "L": The luma (RGB and grayscale images).
        - "L*": The CIE lightness L* (RGB, grayscale, CIELab and CIELuv images).
        - "c*": The CIE chroma c* (CIELab and CIELuv images).
        - "s*": The CIE saturation s* (CIELuv images).
        - "h*": The CIE hue angle h* (CIELab and CIELuv images).

        If it ends with a "+", channels gets appended with the keys already registered in self.stats.
        Default (if None) is "RGBL" for RGB images, "VS" for HSV images, "L'S'" for HSL images,
        "L" for grayscale images, "L*c*" for CIELab and "L*s*" for CIELuv images.
      exclude01 (bool, optional): If True, exclude pixels <= 0 or >= 1 from the median and percentiles.
        Defaults to `equimage.params.exclude01` if None.
      recompute (bool, optional): If False (default), the statistics already registered in self.stats
        are not recomputed. If True, all statistics are recomputed.

    Returns:
      dict: stats[key] for key in channels, with:

        - stats[key].name = channel name (provided for convenience).
        - stats[key].width = image width (provided for convenience).
        - stats[key].height = image height (provided for convenience).
        - stats[key].npixels = number of image pixels = image width*image height (provided for
          convenience).
        - stats[key].minimum = minimum level.
        - stats[key].maximum = maximum level.
        - stats[key].percentiles = (pr25, pr50, pr75) = the 25th, 50th and 75th percentiles.
        - stats[key].median = pr50 = median level.
        - stats[key].zerocount = number of pixels <= 0.
        - stats[key].outcount = number of pixels > 1 (out-of-range).
        - stats[key].exclude01 = True if pixels >= 0 or <= 1 have been excluded from the median and
          percentiles, False otherwise.
        - stats[key].color = suggested text color for display.
    """
    epsilon = helpers.fpepsilon(self.dtype)
    if channels is None:
      if self.colorspace == "CIELab":
        channels = "L*c*"
      elif self.colorspace == "CIELuv":
        channels = "L*s*"
      elif self.colorspace == "lRGB" or self.colorspace == "sRGB":
        if self.colormodel == "RGB":
          channels = "RGBL"
        elif self.colormodel == "HSV":
          channels = "VS"
        elif self.colormodel == "HSL":
          channels = "V'S'"
        elif self.colormodel == "gray":
          channels = "L"
        else:
          self.color_model_error()
      else:
        self.color_space_error()
    if exclude01 is None: exclude01 = params.exclude01
    if not hasattr(self, "stats"): self.stats = {} # Register empty statistics in the object, if none already computed.
    if channels and channels[-1] == "+":
      keys = parse_channels(channels[:-1])
      for key in self.hists.keys(): # Add missing keys.
        if not key in keys: keys.append(key)
    else:
      keys = parse_channels(channels)
    width, height = self.get_size()
    npixels = width*height
    stats = {}
    for key in keys:
      if key in stats: # Already selected.
        print(f"Warning, channel '{key}' selected twice or more...")
        continue
      if not recompute and key in self.stats: # Already computed.
        if self.stats[key].exclude01 == exclude01:
          stats[key] = self.stats[key]
          continue
      if key == "1":
        name = "Channel #1"
        color = "red"
        channel = self.image[0]
      elif key == "2":
        name = "Channel #2"
        color = "green"
        if self.get_nc() < 2: raise self.color_model_error()
        channel = self.image[1]
      elif key == "3":
        name = "Channel #3"
        color = "blue"
        if self.get_nc() < 3: raise self.color_model_error()
        channel = self.image[2]
      elif key == "R":
        self.check_color_model("RGB", "gray")
        name = "Red"
        color = "red"
        channel = self.image[0]
      elif key == "G":
        self.check_color_model("RGB", "gray")
        name = "Green"
        color = "green"
        channel = self.image[1] if self.colormodel == "RGB" else self.image[0]
      elif key == "B":
        self.check_color_model("RGB", "gray")
        name = "Blue"
        color = "blue"
        channel = self.image[2] if self.colormodel == "RGB" else self.image[0]
      elif key == "V":
        name = "HSV value"
        color = "darkslategray"
        channel = self.HSV_value()
      elif key == "S":
        name = "HSV saturation"
        color = "orange"
        channel = self.HSV_saturation()
      elif key == "L'":
        name = "HSL lightness"
        color = "darkslategray"
        channel = self.HSL_lightness()
      elif key == "S'":
        name = "HSL saturation"
        color = "orange"
        channel = self.HSL_saturation()
      elif key == "H":
        name = "HSX hue"
        color = "limegreen"
        channel = self.HSX_hue()
      elif key == "L":
        name = "Luma"
        color = "lightslategray"
        channel = self.luma()
      elif key == "L*":
        name = "Lightness"
        color = "lightsteelblue"
        channel = self.lightness()
      elif key == "c*":
        name = "CIE chroma"
        color = "orange"
        channel = self.CIE_chroma()
      elif key == "s*":
        name = "CIE saturation"
        color = "orange"
        channel = self.CIE_saturation()
      elif key == "h*":
        name = "CIE hue"
        color = "limegreen"
        channel = self.CIE_hue()
      else:
        raise ValueError(f"Error, unknown channel '{key}'.")
      stats[key] = helpers.Container()
      stats[key].name = name
      stats[key].width = width
      stats[key].height = height
      stats[key].npixels = npixels
      stats[key].minimum = np.min(channel)
      stats[key].maximum = np.max(channel)
      if exclude01:
        mask = (channel >= epsilon) & (channel < 1.)
        stats[key].percentiles = np.percentile(channel[mask], [25., 50., 75.]) if np.any(mask) else None
      else:
        stats[key].percentiles = np.percentile(channel, [25., 50., 75.])
      stats[key].median = stats[key].percentiles[1] if stats[key].percentiles is not None else None
      stats[key].zerocount = np.sum(channel < epsilon)
      stats[key].outcount = np.sum(channel > 1.)
      stats[key].exclude01 = exclude01
      stats[key].color = color
    self.stats = stats
    return stats



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Color management.

The following symbols are imported in the equimage/equimagelab namespaces for convenience:
  "HSV_wheel".
"""

__all__ = ["HSV_wheel"]

import os
import numpy as np
import scipy.interpolate as spint
from scipy.interpolate import Akima1DInterpolator

from . import params
from . import helpers
from . import image_colorspaces as colorspaces
from . import image_io as io
from . import __packagepath__

##################
# Test image(s). #
##################

def HSV_wheel():
  """Return a HSV wheel as an Image object, to test color transformations.

  Returns:
    Image: An image object with a HSV wheel.
  """
  # HSVwheel.png was generated by misc/HSVwheel.py.
  image, meta = io.load_image(os.path.join(__packagepath__, "images", "HSVwheel.png"), verbose = False)
  return image

#####################
# Helper functions. #
#####################

def parse_hue_kwargs(D, kwargs):
  """Parse hue keywords in the kwargs.

  This function looks for the keywords 'R' (red, hue H = 0), 'Y' (yellow, H = 1/6), 'G' (green,
  H = 1/3), 'C' (cyan, H = 1/2), 'B' (blue, H = 2/3) and 'M' (magenta, H = 5/6) in the kwargs and
  returns the grid of H's and the corresponding values of the kwargs as numpy arrays. Whenever a
  keyword is missing, its value is replaced by the default, D.
  Additional points may be inserted in the grid by providing the keywords 'RY' (red-yellow, H = 1/12),
  'YG' (yellow-green, H = 1/4), 'GC' (green-cyan, H = 5/12), 'CB' (cyan-blue, H = 7/12), 'BM' (blue-
  magenta, H = 3/4) and 'MR' (magenta-red, H = 11/12).

  Note:
    Used by :meth:`Image.HSX_color_saturation() <.HSX_color_saturation>`, :meth:`Image.CIE_chroma_saturation() <.CIE_chroma_saturation>`,
    :meth:`Image.rotate_HSX_hue() <.rotate_HSX_hue>`, :meth:`Image.rotate_CIE_hue() <.rotate_CIE_hue>`.

  Args:
    D (float): The default value for the R/Y/G/C/B/M hues.
    kwargs (dict): The dictionary of kwargs.

  Returns:
    The grid of hues (numpy.ndarray), the corresponding keyword values (numpy.ndarray), and the
    curated kwargs (with the used keys deleted).
  """
  R = kwargs.pop("R", D)
  Y = kwargs.pop("Y", D)
  G = kwargs.pop("G", D)
  C = kwargs.pop("C", D)
  B = kwargs.pop("B", D)
  M = kwargs.pop("M", D)
  hgrid = [0., 1./6., 2./6., 3./6., 4./6., 5./6., 1.] # Enforce periodic boundary conditions.
  value = [R, Y, G, C, B, M, R]
  if (RY := kwargs.pop("RY", None)) is not None:
    hgrid.append(1./12.)
    value.append(RY)
  if (YG := kwargs.pop("YG", None)) is not None:
    hgrid.append(3./12.)
    value.append(YG)
  if (GC := kwargs.pop("GC", None)) is not None:
    hgrid.append(5./12.)
    value.append(GC)
  if (CB := kwargs.pop("CB", None)) is not None:
    hgrid.append(7./12.)
    value.append(CB)
  if (BM := kwargs.pop("BM", None)) is not None:
    print(BM)
    hgrid.append(9./12.)
    value.append(BM)
  if (MR := kwargs.pop("MR", None)) is not None:
    hgrid.append(11./12.)
    value.append(MR)
  hgrid = np.array(hgrid)
  value = np.array(value)
  idx = np.argsort(hgrid)
  return hgrid[idx], value[idx], kwargs

def interpolate_hue(hue, hgrid, param, interpolation):
  """Interpolate a parameter param defined on a grid of hues.

  Note:
    Used by :meth:`Image.HSX_color_saturation() <.HSX_color_saturation>`, :meth:`Image.CIE_chroma_saturation() <.CIE_chroma_saturation>`,
    :meth:`Image.rotate_HSX_hue() <.rotate_HSX_hue>`, :meth:`Image.rotate_CIE_hue() <.rotate_CIE_hue>`.

  Args:
    hue (numpy.ndarray): The input hues at which the parameter must be interpolated.
    hgrid (numpy.ndarray): The grid of hues on which the parameter is defined.
    param (numpy.ndarray): The parameter on the grid of hues.
    interpolation (str, optional): The interpolation method:

      - "nearest": Nearest neighbor interpolation.
      - "linear": Linear spline interpolation.
      - "cubic": Cubic spline interpolation.
      - "akima": Akima spline interpolation (default).

  Returns:
    numpy.ndarray: The parameter interpolated for all input hues.
  """
  if np.all(param == param[0]):
    return np.full_like(hue, param[0]) # Short-cut if the parameter is the same for all points of the grid.
  if interpolation == "nearest":
    finter = spint.interp1d(hgrid, param, kind = "nearest")
  elif interpolation == "linear" or interpolation == "cubic":
    k = 3 if interpolation == "cubic" else 1
    tck = spint.splrep(hgrid, param, k = k, per = True)
    def finter(x): spint.splev(x, tck)
  elif interpolation == "akima":
    hgrid = np.concatenate(([hgrid[-3]-1., hgrid[-2]-1.], hgrid, [hgrid[1]+1., hgrid[2]+1.])) # Enforce periodic boundary conditions.
    param = np.concatenate(([param[-3], param[-2]], param, [param[1], param[2]]))
    finter = Akima1DInterpolator(hgrid, param)
  else:
    raise ValueError(f"Error, unknown interpolation method '{interpolation}'.")
  return finter(hue)

#####################################
# For inclusion in the Image class. #
#####################################

class MixinImage:
  """To be included in the Image class."""

  ##################
  # Image queries. #
  ##################

  def is_grayscale_RGB(self):
    """Return True if a RGB image is actually a grayscale (same RGB channels), False otherwise.

    Returns:
      bool: True if the RGB image is a grayscale (same RGB channels), False otherwise.
    """
    self.check_color_model("RGB")
    epsilon = helpers.fpepsilon(self.dtype)
    return np.all(abs(self.image[1]-self.image[0]) < epsilon) and np.all(abs(self.image[2]-self.image[0]) < epsilon)

  ############################
  # Gray scales & negatives. #
  ############################

  def negative(self):
    """Return the negative of a RGB or grayscale image.

    Returns:
      Image: The negative of the image.
    """
    self.check_color_model("RGB", "gray")
    return 1.-self

  def grayscale(self, channel = "L*", RGB = False):
    """Convert the selected channel of a RGB image into a grayscale image.

    Args:
      channel: The converted channel ("V" for the HSV value, "L'" for HSL lightness, "L" for the
        luma, "Y" or "L*" for the luminance/lightness). Namely, the output grayscale image has the
        same value/luma/luminance and lightness as the original RGB image.
      RGB (bool, optional): If True, return the grayscale as a RGB image (with identical R/G/B
        channels). If False (default), return the grayscale as a single channel image.

    Returns:
      Image: The grayscale image.
    """
    self.check_color_model("RGB")
    if channel == "V":
      grayscale = self.HSV_value()
    elif channel == "L'":
      grayscale = self.HSL_lightness()
    elif channel == "L":
      grayscale = self.luma()
    elif channel == "Y" or channel == "L*":
      grayscale = self.luminance()
      if self.colorspace == "sRGB":
        grayscale = colorspaces.lRGB_to_sRGB(grayscale) # Preserve luminance & lightness in the sRGB color space.
    else:
      raise ValueError(f"Error, invalid channel '{channel}' (must be 'V', 'L' or 'Y'/'L*').")
    if RGB:
      return self.newImage(np.repeat(grayscale[np.newaxis, :, :], 3, axis = 0))
    else:
      return self.newImage(grayscale, colormodel = "gray")

  ##########################
  # Color transformations. #
  ##########################

  def neutralize_background(self, source, neutral = None, mode = "additive"):
    """Neutralize the background of a RGB image (turn a background color into gray).

    Given a source background color (Rs, Gs, Bs), and a target neutral level N, this method
    transforms the RGB channels as:

      - R ← R-Rs+N
      - G ← G-Gs+N
      - B ← B-Bs+N

    if mode = "additive", or as:

      - R ← R*N/Rs
      - G ← G*N/Gs
      - B ← B*N/Bs

    if mode = "multiplicative", with N = max(Rs, Gs, Bs) by default. On output, the source
    background color becomes, therefore, the gray (N, N, N) color in both cases.

    Args:
      source (float): The source background color (tuple/list/array of the Rs, Gs, Bs levels).
      neutral (float, optional): The target neutral level [max(Rs, Gs, Bs) if None (default)].
      mode (str, optional): The neutralization mode ["additive" (default) or "multiplicative"].

    Returns:
      Image: The processed image.
    """
    self.check_color_model("RGB")
    source = np.asarray(source)
    if source.shape != (3,): raise ValueError("Error, source must be a tuple/list/array of three elements.")
    source = source.reshape((3, 1, 1))
    if neutral is None: neutral = source.max()
    if mode == "additive":
      return self+(neutral-source)
    elif mode == "multiplicative":
      return self*(neutral/source)
    else:
      raise ValueError("Error, mode must be 'additive' or 'multiplicative'.")

  def color_balance(self, red = 1., green = 1., blue = 1., neutral = 0.):
    """Adjust the color balance of a RGB image.

    This method linearly scales the RGB channels as:

      - R ← red*(R-neutral)+neutral
      - G ← green*(G-neutral)+neutral
      - B ← blue*(B-neutral)+neutral

    The neutral color thus remains unchanged.

    Args:
      red (float, optional): The multiplier for the red channel (default 1).
      green (float, optional): The multiplier for the green channel (default 1).
      blue (float, optional): The multiplier for the blue channel (default 1).
      neutral (float, optional): The neutral level or color (default 0).
        Can be a scalar or a tuple/list/array of neutral (R, G, B) levels.

    Returns:
      Image: The processed image.
    """
    self.check_color_model("RGB")
    if red < 0.: raise ValueError("Error, red must be >= 0.")
    if green < 0.: raise ValueError("Error, green must be >= 0.")
    if blue < 0.: raise ValueError("Error, blue must be >= 0.")
    mult = np.array([[[red]], [[green]], [[blue]]])
    neutral = np.asarray(neutral)
    if neutral.ndim == 0: neutral = np.array([neutral, neutral, neutral])
    if neutral.shape != (3,): raise ValueError("Error, neutral must be a scalar or a tuple/list/array of three elements.")
    neutral = neutral.reshape((3, 1, 1))
    return mult*(self-neutral)+neutral

  def match_RGB(self, source, target, neutral = 0.):
    """Adjust the color balance of a RGB image to transform a source into a target color.

    Given a source color (Rs, Gs, Bs), a target color (Rt, Gt, Bt), and a neutral level N, this
    method linearly scales the RGB channels as:

      - R ← (Rt-N)/(Rs-N)*(R-N)+N
      - G ← (Gt-N)/(Gs-N)*(G-N)+N
      - B ← (Bt-N)/(Bs-N)*(B-N)+N

    On output, the source color becomes the target color, while the neutral color remains unchanged.

    Args:
      source (float): The source color (tuple/list/array of the Rs, Gs, Bs levels).
      target (float): The target color (tuple/list/array of the Rt, Gt, Bt levels).
      neutral (float, optional): The neutral level or color (default 0).
        Can be a scalar or a tuple/list/array of neutral (R, G, B) levels.

    Returns:
      Image: The processed image.
    """
    source = np.asarray(source)
    if source.shape != (3,): raise ValueError("Error, source must be a tuple/list/array of three elements.")
    source = source.reshape((3, 1, 1))
    target = np.asarray(target)
    if target.shape != (3,): raise ValueError("Error, target must be a tuple/list/array of three elements.")
    target = target.reshape((3, 1, 1))
    neutral = np.asarray(neutral)
    if neutral.ndim == 0: neutral = np.array([neutral, neutral, neutral])
    if neutral.shape != (3,): raise ValueError("Error, neutral must be a scalar or a tuple/list/array of three elements.")
    neutral = neutral.reshape((3, 1, 1))
    mult = (target-neutral)/(source-neutral)
    return mult*(self-neutral)+neutral

  def mix_RGB(self, M, neutral = 0.):
    """Mix RGB channels.

    Transforms each pixel P = (R, G, B) of the image as:

      P ← M@(P-neutral)+neutral,

    with M a 3x3 mixing matrix. The neutral color thus remains unchanged.

    Args:
      M (numpy.ndarray): The mixing matrix.
      neutral (float, optional): The neutral level or color (default 0).
        Can be a scalar or a tuple/list/array of neutral (R, G, B) levels.

    Returns:
      Image: The processed image.
    """
    self.check_color_model("RGB")
    M = np.asarray(M)
    if M.shape != (3, 3): raise ValueError("Error, M must be a 3x3 matrix.")
    neutral = np.asarray(neutral)
    if neutral.ndim == 0: neutral = np.array([neutral, neutral, neutral])
    if neutral.shape != (3,): raise ValueError("Error, neutral must be a scalar or a tuple/list/array of three elements.")
    neutral = neutral.reshape((3, 1, 1))
    return self.newImage(np.tensordot(M, self.image-neutral, axes = 1)+neutral)

  def color_temperature(self, T, T0 = 6650., lightness = False):
    """Adjust the color temperature of a RGB image.

    Adjusts the color balance assuming that the scene is (or is lit by) a black body source whose
    temperature is changed from T0 (default 6650K) to T.
    Setting T < T0 casts a red tint on the image, while setting T > T0 casts a blue tint.
    This is not a rigorous transformation and is intended for "cosmetic" purposes.
    The colors are balanced in the linear RGB color space.

    Args:
      T (float): The target temperature between 1000K and 40000K.
      T0 (float, optional): The initial temperature between 1000K and 40000K (default 6650K).
      lightness (bool, optional): If True, preserve the lightness L* of the original image.
        Note that this may result in some out-of-range pixels. Default is False.

    Returns:
      Image: The processed image.
    """

    def RGB_multipliers(T):
      """Return RGB multipliers for a given temperature T.

      These are fits to the data of:
      http://www.vendian.org/mncharity/dir3/blackbody/
      http://www.vendian.org/mncharity/dir3/blackbody/UnstableURLs/bbr_color.html
      See misc/CTdata.py for the fits.

      Args:
        T (float): The temperature (K).

      Returns:
        float: The red, green, blue multipliers for temperature T.
      """

      def fit_function(T, a, b, c, n):
        """Fit function for the RGB multipliers."""
        Tr = (T-6650.)/10000.
        return max(0., 1.+a*Tr+b*Tr**2+c*Tr**3)**n

      if T < 1000. or T > 40000.:
        raise ValueError("Error, the temperature must range between 1000K and 40000K.")
      if T < 6650.:
        red = 1.
        green = .955*fit_function(T, 0.6328063016856568, -0.7554188937494166, 1.7905208976761535, 1.2444903530611828)
        blue = fit_function(T, 1.2773641159953395, -1.0621459476222992, 1.1594787131985838, 1.7962669751286322)
      else:
        red = fit_function(T, 3.7874055198445142, -1.2956261274808025, 0.16818980687443094, -0.6136402157590004)
        green = .955*fit_function(T, 3.490058031650525, -1.1425871791667106, 0.14401832008759521, -0.42381907464549073)
        blue = 1.
      return red, green, blue

    self.check_color_space("lRGB", "sRGB")
    red0, green0, blue0 = RGB_multipliers(T0)
    red , green , blue  = RGB_multipliers(T)
    red /= red0 ; green /= green0 ; blue /= blue0
    maxi = np.max([red, green, blue])
    red /= maxi ; green /= maxi ; blue /= maxi
    print(f"Red multiplier = {red:.3f}.")
    print(f"Green multiplier = {green:.3f}.")
    print(f"Blue multiplier = {blue:.3f}.")
    image = self.convert(colorspace = "lRGB", colormodel = "RGB", copy = False)
    balanced = image.color_balance(red, green, blue)
    if lightness: balanced.set_channel("L*sh", self.lightness(), inplace = True)
    return balanced.convert(colorspace = self.colorspace, colormodel = self.colormodel, copy = False)

  def HSX_color_saturation(self, D = 0., mode = "midsat", colormodel = "HSV", colorspace = None, interpolation = "akima", lightness = False, trans = True, **kwargs):
    """Adjust color saturation in the HSV or HSL color models.

    The image is converted (if needed) to the HSV or HSL color model, then the color saturation S is
    transformed according to the mode kwarg:

      - "addsat": Shift the saturation S ← S+delta.
      - "mulsat": Scale the saturation S ← S*(1+delta).
      - "midsat": Apply a midtone stretch function S ← f(S) = (m-1)S/((2m-1)S-m) with midtone m = (1-delta)/2.
        This function increases monotonously from f(0) = 0 to f(m) = 1/2 and f(1) = 1, and thus leaves
        the saturation of the least/most saturated pixels unchanged.

    The image is then converted back to the original color model after this operation.
    delta is expected to be > -1, and to be < 1 in the "midsat" mode. Whatever the mode, delta = 0
    leaves the image unchanged, delta > 0 saturates the colors, and delta < 0 turns the image into
    a gray scale. delta is set for the red ('R'), yellow ('Y'), green ('G'), cyan ('C'), blue ('B')
    and magenta ('M') hues by the corresponding kwarg (delta = D if missing). It is interpolated
    for arbitrary hues using nearest neighbor, linear, cubic or akima spline interpolation according
    to the interpolation kwarg. Midpoint deltas may also be specified for finer interpolation by
    providing the kwargs 'RY' (red-yellow), 'YG' (yellow-green), 'GC' (green-cyan), 'CB' (cyan-blue),
    'BM' (blue-magenta) and 'MR' (magenta-red).

    See also:
      :meth:`Image.CIE_chroma_saturation() <.CIE_chroma_saturation>`

    Args:
      D (float, optional): The delta for all hues (default 0).
      R (float, optional): The red delta (default D).
      Y (float, optional): The yellow delta (default D).
      G (float, optional): The green delta (default D).
      C (float, optional): The cyan delta (default D).
      B (float, optional): The blue delta (default D).
      M (float, optional): The magenta delta (default D).
      mode (str, optional): The saturation mode ["addsat", "mulsat" or "midsat" (default)].
      colormodel (str, optional): The color model for saturation ["HSV" (default) or "HSL"].
      colorspace (str, optional): The color space for saturation ["lRGB", "sRGB", or None (default)
        to use the color space of the image].
      interpolation (str, optional): The interpolation method for delta(hue):

        - "nearest": Nearest neighbor interpolation.
        - "linear": Linear spline interpolation.
        - "cubic": Cubic spline interpolation.
        - "akima": Akima spline interpolation (default).

      lightness (bool, optional): If True, preserve the lightness L* of the original image.
        Note that this may result in some out-of-range pixels. Default is False.
      trans (bool, optional): If True (default), embed the transormation in the output image
        as output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The processed image.
    """
    self.check_color_space("lRGB", "sRGB")
    hgrid, dgrid, kwargs = parse_hue_kwargs(D, kwargs)
    if kwargs: raise ValueError(f"Error, unknown keyword argument(s): {', '.join(kwargs.keys())}.")
    if colorspace is None:
      colorspace = self.colorspace
    elif colorspace not in ["lRGB", "sRGB"]:
      raise ValueError("Error, colorspace must be 'lRGB' or 'sRGB'.")
    if colormodel == "HSV":
      channel = "S"
    elif colormodel == "HSL":
      channel = "S'"
    else:
      raise ValueError("Error, colormodel must be 'HSV' or 'HSL'.")
    hsx = self.convert(colorspace = colorspace, colormodel = colormodel, copy = True)
    hue = hsx.image[0]
    sat = hsx.image[1]
    print(f"Before operation: min(saturation) = {np.min(sat):.5f}; median(saturation) = {np.median(sat):.5f}; max(saturation) = {np.max(sat):.5f}.")
    delta = interpolate_hue(hue, hgrid, dgrid, interpolation)
    if mode == "addsat":
      sat += delta
    elif mode == "mulsat":
      sat *= 1.+delta
    elif mode == "midsat":
      midsat = np.clip(.5*(1.-delta), .001, .999)
      sat = (midsat-1.)*sat/((2.*midsat-1.)*sat-midsat)
    else:
      raise ValueError(f"Error, unknown saturation mode '{mode}.")
    print(f"After  operation: min(saturation) = {np.min(sat):.5f}; median(saturation) = {np.median(sat):.5f}; max(saturation) = {np.max(sat):.5f}.")
    hsx.image[1] = np.clip(sat, 0., 1.)
    output = hsx.convert(colorspace = self.colorspace, colormodel = self.colormodel, copy = False)
    if lightness: output.set_channel("L*sh", self.lightness(), inplace = True)
    if trans:
      t = helpers.Container()
      t.type = "hue"
      t.input = self
      t.channels = channel if self.colorspace == colorspace else ""
      t.xm = np.linspace(0., 1., 7)
      t.ym = interpolate_hue(t.xm, hgrid, dgrid, interpolation)
      t.cm = ["red", "yellow", "green", "cyan", "blue", "magenta", "red"]
      t.x = np.linspace(0., 1., params.ntranslo)
      t.y = interpolate_hue(t.x, hgrid, dgrid, interpolation)
      t.ylabel = "\u0394"
      output.trans = t
    return output

  def CIE_chroma_saturation(self, D = 0., mode = "midsat", colormodel = "Lsh", interpolation = "akima", ref = None, trans = True, **kwargs):
    """Adjust color chroma or saturation in the CIELab or CIELuv color spaces.

    The image is converted (if needed) to the CIELab or CIELuv colorspace, then the CIELab chroma
    CS = c* = sqrt(a*^2+b*^2) (colormodel = "Lab"), or the CIELuv chroma CS = c* = sqrt(u*^2+v*^2)
    (colormodel = "Luv"), or the CIELuv saturation CS = s* = c*/L* (colormodel = "Lsh") is transformed
    according to the mode kwarg:

      - "addsat": Shift the chroma/saturation CS ← CS+delta.
      - "mulsat": Scale the chroma/saturation CS ← CS*(1+delta).
      - "midsat": Apply a midtone stretch function CS ← f(CS) = (m-1)CS/((2m-1)CS/ref-m) with midtone m = (1-delta)/2.
        This function increases monotonously from f(0) = 0 to f(m*ref) = ref/2 and f(ref) = ref,
        where ref is a reference chroma/saturation (ref = max(CS) by default).

    The image is then converted back to the original color space and model after this operation.
    delta is expected to be > -1, and to be < 1 in the "midsat" mode. Whatever the mode, delta = 0
    leaves the image unchanged, delta > 0 saturates the colors, and delta < 0 turns the image into
    a gray scale. However, please keep in mind that the chroma/saturation in the CIELab/CIELuv color
    spaces is not bounded by 1 as it is in the lRGB and sRGB color spaces (HSV and HSL color models).
    The choice of the reference can, therefore, be critical in the "midsat" mode. In particular,
    pixels with chroma/saturation > ref get desaturated if delta > 0, and oversaturated if delta < 0
    (with a possible singularity at CS = -ref*(1-delta)/(2*delta)).
    delta is set for the red ('R'), yellow ('Y'), green ('G'), cyan ('C'), blue ('B') and magenta
    ('M') hues by the corresponding kwarg (delta = D if missing). It is interpolated for arbitrary
    hues using nearest neighbor, linear, cubic or akima spline interpolation according to the
    interpolation kwarg. Midpoint deltas may also be specified for finer interpolation by providing
    the kwargs 'RY' (red-yellow), 'YG' (yellow-green), 'GC' (green-cyan), 'CB' (cyan-blue), 'BM'
    (blue-magenta) and 'MR' (magenta-red).
    Contrary to the saturation of HSV or HSL images, chroma/saturation transformations in the CIELab
    and CIELuv color spaces preserve the lightness by design. They may, however, result in out-of-
    range RGB pixels (as not all points of of these color spaces correspond to physical RGB colors).

    Note:
      Chroma and saturation are related, but different quantities (s* = c*/L* in the CIELuv color space).
      There is no rigorous definition of saturation in the CIELab color space.

    See also:
      :meth:`Image.HSX_color_saturation() <.HSX_color_saturation>`

    Args:
      D (float, optional): The delta for all hues (default 0).
      R (float, optional): The red delta (default D).
      Y (float, optional): The yellow delta (default D).
      G (float, optional): The green delta (default D).
      C (float, optional): The cyan delta (default D).
      B (float, optional): The blue delta (default D).
      M (float, optional): The magenta delta (default D).
      mode (str, optional): The saturation mode ["addsat", "mulsat" or "midsat" (default)].
      colormodel (str, optional): The color model for saturation ["Lab", "Luv" or "Lsh" (default)].
      interpolation (str, optional): The interpolation method for delta(hue angle):

        - "nearest": Nearest neighbor interpolation.
        - "linear": Linear spline interpolation.
        - "cubic": Cubic spline interpolation.
        - "akima": Akima spline interpolation (default).

      ref (float, optional): The reference chroma/saturation for the "midsat" mode. If None,
        defaults to the maximum chroma/saturation of the input image.
      trans (bool, optional): If True (default), embed the transormation in the output image
        as output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The processed image.
    """
    hgrid, dgrid, kwargs = parse_hue_kwargs(D, kwargs)
    if kwargs: raise ValueError(f"Error, unknown keyword argument(s): {', '.join(kwargs.keys())}.")
    if colormodel == "Lab":
      name = "chroma"
      channel = "c*" if self.colorspace == "CIELab" else ""
      CIE = self.convert(colorspace = "CIELab", colormodel = "Lch", copy = True)
    elif colormodel == "Luv":
      name = "chroma"
      channel = "c*" if self.colorspace == "CIELuv" else ""
      CIE = self.convert(colorspace = "CIELuv", colormodel = "Lch", copy = True)
    elif colormodel == "Lsh":
      name = "saturation"
      channel = "s*" if self.colorspace == "CIELuv" else ""
      CIE = self.convert(colorspace = "CIELuv", colormodel = "Lsh", copy = True)
    else:
      raise ValueError("Error, colormodel must be 'Lab' or 'Luv' or 'Lsh'.")
    hue = CIE.image[2]
    sat = CIE.image[1]
    maxsat = np.max(sat)
    print(f"Before operation: min({name}) = {np.min(sat):.5f}; median({name}) = {np.median(sat):.5f}; max({name}) = {maxsat:.5f}.")
    delta = interpolate_hue(hue, hgrid, dgrid, interpolation)
    if mode == "addsat":
      sat += delta
    elif mode == "mulsat":
      sat *= 1.+delta
    elif mode == "midsat":
      if ref is None: ref = maxsat
      midsat = np.clip(.5*(1.-delta), .001, .999)
      sat = (midsat-1.)*sat/((2.*midsat-1.)*sat/ref-midsat)
    else:
      raise ValueError(f"Error, unknown saturation mode '{mode}.")
    print(f"After  operation: min({name}) = {np.min(sat):.5f}; median({name}) = {np.median(sat):.5f}; max({name}) = {np.max(sat):.5f}.")
    CIE.image[1] = sat
    output = CIE.convert(colorspace = self.colorspace, colormodel = self.colormodel, copy = False)
    if trans:
      t = helpers.Container()
      t.type = "hue"
      t.input = self
      t.channels = channel
      t.xm = np.linspace(0., 1., 7)
      t.ym = interpolate_hue(t.xm, hgrid, dgrid, interpolation)
      t.cm = ["red", "yellow", "green", "cyan", "blue", "magenta", "red"]
      t.x = np.linspace(0., 1., params.ntranslo)
      t.y = interpolate_hue(t.x, hgrid, dgrid, interpolation)
      t.ylabel = "\u0394"
      output.trans = t
    return output

  def rotate_HSX_hue(self, D = 0., colorspace = None, interpolation = "akima", lightness = False, trans = True, **kwargs):
    """Rotate color hues in the HSV/HSL color models.

    The image is converted (if RGB) to the HSV color model, and the hue H is rotated:

      H ← (H+delta)%1.

    The image is then converted back to the original color model after this operation.
    delta is set for the original red ('R'), yellow ('Y'), green ('G'), cyan ('C'), blue ('B') and
    magenta ('M') hues by the corresponding kwarg (delta = D if missing). It is interpolated for
    arbitrary hues using nearest neighbor, linear, cubic or akima spline interpolation according to
    the interpolation kwarg. Midpoint deltas may also be specified for finer interpolation by
    providing the kwargs 'RY' (red-yellow), 'YG' (yellow-green), 'GC' (green-cyan), 'CB' (cyan-blue),
    'BM' (blue-magenta) and 'MR' (magenta-red).

    Note:
      H(red) = 0, H(yellow) = 1/6, H(green) = 1/3, H(cyan) = 1/2, H(blue) = 2/3, and H(magenta) = 5/6.
      A uniform rotation D = 1/6 converts red → yellow, yellow → green, green → cyan, cyan → blue,
      blue → magenta, and magenta → red.
      A uniform rotation D = -1/6 converts red → magenta, yellow → red, green → yellow, cyan → green,
      blue → cyan, and magenta → blue.

    See also:
      :meth:`Image.rotate_CIE_hue() <.rotate_CIE_hue>`

    Args:
      D (float, optional): The delta for all hues (default 0).
      R (float, optional): The red delta (default D).
      Y (float, optional): The yellow delta (default D).
      G (float, optional): The green delta (default D).
      C (float, optional): The cyan delta (default D).
      B (float, optional): The blue delta (default D).
      M (float, optional): The magenta delta (default D).
      colorspace (str, optional): The color space for saturation ["lRGB", "sRGB", or None (default)
        to use the color space of the image].
      interpolation (str, optional): The interpolation method for delta(hue):

        - "nearest": Nearest neighbor interpolation.
        - "linear": Linear spline interpolation.
        - "cubic": Cubic spline interpolation.
        - "akima": Akima spline interpolation (default).

      lightness (bool, optional): If True, preserve the lightness L* of the original image.
        Note that this may result in some out-of-range pixels. Default is False.
      trans (bool, optional): If True (default), embed the transormation in the output image
        as output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The processed image.
    """
    self.check_color_space("lRGB", "sRGB")
    hgrid, dgrid, kwargs = parse_hue_kwargs(D, kwargs)
    if kwargs: raise ValueError(f"Error, unknown keyword argument(s): {', '.join(kwargs.keys())}.")
    if colorspace is None:
      colorspace = self.colorspace
    elif colorspace not in ["lRGB", "sRGB"]:
      raise ValueError("Error, colorspace must be 'lRGB' or 'sRGB'.")
    colormodel = self.colormodel if self.colormodel in ["HSV", "HSL"] else "HSV"
    hsx = self.convert(colorspace = colorspace, colormodel = colormodel, copy = True)
    hue = hsx.image[0]
    delta = interpolate_hue(hue, hgrid, dgrid, interpolation)
    hsx.image[0] = (hue+delta)%1.
    output = hsx.convert(colorspace = self.colorspace, colormodel = self.colormodel, copy = False)
    if lightness: output.set_channel("L*sh", self.lightness(), inplace = True)
    if trans:
      t = helpers.Container()
      t.type = "hue"
      t.input = self
      t.channels = "H" if self.colorspace == colorspace else ""
      t.xm = np.linspace(0., 1., 7)
      t.ym = interpolate_hue(t.xm, hgrid, dgrid, interpolation)
      t.cm = ["red", "yellow", "green", "cyan", "blue", "magenta", "red"]
      t.x = np.linspace(0., 1., params.ntranslo)
      t.y = interpolate_hue(t.x, hgrid, dgrid, interpolation)
      t.ylabel = "\u0394"
      output.trans = t
    return output

  def rotate_CIE_hue(self, D = 0., colorspace = "CIELab", interpolation = "akima", trans = True, **kwargs):
    """Rotate color hues in the CIELab or CIELuv color space.

    The image is converted (if needed) to the CIELab or CIELuv color space, and the reduced hue
    angle h* (within [0, 1]) is rotated:

      h* ← (h*+delta)%1.

    The image is then converted back to the original color model after this operation.
    delta is set for the original red ('R'), yellow ('Y'), green ('G'), cyan ('C'), blue ('B') and
    magenta ('M') hues by the corresponding kwarg (delta = D if missing). It is interpolated for
    arbitrary hues using nearest neighbor, linear, cubic or akima spline interpolation according to
    the interpolation kwarg. Midpoint deltas may also be specified for finer interpolation by
    providing the kwargs 'RY' (red-yellow), 'YG' (yellow-green), 'GC' (green-cyan), 'CB' (cyan-blue),
    'BM' (blue-magenta) and 'MR' (magenta-red).
    Contrary to the rotation of HSV or HSL images, rotations in the CIELab and CIELuv color spaces
    preserve the lightness by design. They may, however, result in out-of-range RGB pixels (as not
    all points of of these color spaces correspond to physical RGB colors).

    Note:
      h*(red) ~ 0, h*(yellow) ~ 1/6, h*(green) ~ 1/3, h*(cyan) ~ 1/2, h*(blue) ~ 2/3, and h*(magenta) ~ 5/6.
      A uniform rotation D = 1/6 converts red → yellow, yellow → green, green → cyan, cyan → blue,
      blue → magenta, and magenta → red.
      A uniform rotation D = -1/6 converts red → magenta, yellow → red, green → yellow, cyan → green,
      blue → cyan, and magenta → blue.

    See also:
      :meth:`Image.rotate_HSX_hue() <.rotate_HSX_hue>`

    Args:
      D (float, optional): The delta for all hues (default 0).
      R (float, optional): The red delta (default D).
      Y (float, optional): The yellow delta (default D).
      G (float, optional): The green delta (default D).
      C (float, optional): The cyan delta (default D).
      B (float, optional): The blue delta (default D).
      M (float, optional): The magenta delta (default D).
      colorspace (str, optional): The color space for rotation ["CIELab" (default) or "CIELuv"].
      interpolation (str, optional): The interpolation method for delta(hue angle):

        - "nearest": Nearest neighbor interpolation.
        - "linear": Linear spline interpolation.
        - "cubic": Cubic spline interpolation.
        - "akima": Akima spline interpolation (default).

      trans (bool, optional): If True (default), embed the transormation in the output image
        as output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The processed image.
    """
    hgrid, dgrid, kwargs = parse_hue_kwargs(D, kwargs)
    if kwargs: raise ValueError(f"Error, unknown keyword argument(s): {', '.join(kwargs.keys())}.")
    if colorspace in ["CIELab", "CIELuv"]:
      CIE = self.convert(colorspace = colorspace, colormodel = "Lch", copy = True)
    else:
      raise ValueError("Error, colorspace must be 'CIELab' or 'CIELuv'.")
    hue = CIE.image[2]
    delta = interpolate_hue(hue, hgrid, dgrid, interpolation)
    CIE.image[2] = (hue+delta)%1.
    output = CIE.convert(colorspace = self.colorspace, colormodel = self.colormodel, copy = False)
    if trans:
      t = helpers.Container()
      t.type = "hue"
      t.input = self
      t.channels = "h*" if self.colorspace in ["CIELab", "CIELuv"] else ""
      t.xm = np.linspace(0., 1., 7)
      t.ym = interpolate_hue(t.xm, hgrid, dgrid, interpolation)
      t.cm = ["red", "yellow", "green", "cyan", "blue", "magenta", "red"]
      t.x = np.linspace(0., 1., params.ntranslo)
      t.y = interpolate_hue(t.x, hgrid, dgrid, interpolation)
      t.ylabel = "\u0394"
      output.trans = t
    return output

  ##########################
  # Color noise reduction. #
  ##########################

  def SCNR(self, hue = "green", protection = "avgneutral", amount = 1., colorspace = None, lightness = True):
    """Subtractive Chromatic Noise Reduction of a given hue of a RGB image.

    The input hue is reduced according to the protection kwarg. For the green hue for example,

      - G ← min(G, C) with C = (R+B)/2 for average neutral protection (protection = "avgneutral").
      - G ← min(G, C) with C = max(R, B) for maximum neutral protection (protection = "maxneutral").
      - G ← G*[(1-A)+C*A] with C = (R+B)/2 for additive mask protection (protection = "addmask").
      - G ← G*[(1-A)+C*A] with C = max(R, B) for maximum mask protection (protection = "maxmask").

    The parameter A in [0, 1] controls the strength of the mask protection.

    Args:
      hue (str, optional): The hue to be reduced ["red" alias "R", "yellow" alias "Y", "green"
        alias "G" (default), "cyan" alias "C", "blue" alias "B", or "magenta" alias "M"].
      protection (str, optional): The protection mode ["avgneutral" (default), "maxneutral",
        "addmask" or "maxmask"].
      amount (float, optional): The parameter A for mask protection (protection = "addmask"
        or "maxmask", default 1).
      colorspace (str, optional): The color space for SCNR ["lRGB", "sRGB", or None (default)
        to use the color space of the image].
      lightness (bool, optional): If True (default), preserve the lightness L* of the original image.
        Note that this may result in some out-of-range pixels.

    Returns:
      Image: The processed image.
    """
    self.check_color_space("lRGB", "sRGB")
    if hue == "red" or hue == "R":
      icc, ic1, ic2, negative = 0, 1, 2, False
    elif hue == "yellow" or hue == "Y":
      icc, ic1, ic2, negative = 2, 0, 1, True
    elif hue == "green" or hue == "G":
      icc, ic1, ic2, negative = 1, 0, 2, False
    elif hue == "cyan" or hue == "C":
      icc, ic1, ic2, negative = 0, 1, 2, True
    elif hue == "blue" or hue == "B":
      icc, ic1, ic2, negative = 2, 0, 1, False
    elif hue == "magenta" or hue == "M":
      icc, ic1, ic2, negative = 1, 0, 2, True
    else:
      raise ValueError(f"Error, unknown hue '{hue}'.")
    if colorspace is None:
      colorspace = self.colorspace
    elif colorspace not in ["lRGB", "sRGB"]:
      raise ValueError("Error, colorspace must be 'lRGB' or 'sRGB'.")
    converted = self.convert(colorspace = colorspace, colormodel = "RGB", copy = False)
    image = np.clip(converted.image, 0., 1.) # Clip before reducing color noise.
    if negative: image = 1.-image
    if protection == "avgneutral":
      c = (image[ic1]+image[ic2])/2.
      image[icc] = np.minimum(image[icc], c)
    elif protection == "maxneutral":
      c = np.maximum(image[ic1], image[ic2])
      image[icc] = np.minimum(image[icc], c)
    elif protection == "addmask":
      c = np.minimum(1., image[ic1]+image[ic2])
      image[icc] *= (1.-amount)+c*amount
    elif protection == "maxmask":
      c = np.maximum(image[ic1], image[ic2])
      image[icc] *= (1.-amount)+c*amount
    else:
      raise ValueError(f"Error, unknown protection mode '{protection}'.")
    if negative: image = 1.-image
    output = converted.newImage(image)
    if lightness: output.set_channel("L*sh", self.lightness(), inplace = True)
    return output.convert(colorspace = self.colorspace, copy = False)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Image masks.

The following symbols are imported in the equimage/equimagelab namespaces for convenience:
  "float_mask", "extend_bmask", "smooth_mask", "threshold_bmask", "threshold_fmask", "shape_bmask".
"""

__all__ = ["float_mask", "extend_bmask", "smooth_mask", "threshold_bmask", "threshold_fmask", "shape_bmask"]

import numpy as np
import scipy.ndimage as ndimg
import skimage.draw as skdraw
import skimage.morphology as skimo

from . import params

####################################
# Binary & float masks management. #
####################################

def float_mask(mask):
  """Convert a binary mask into a float mask.

  Args:
    mask (numpy.ndarray): The input binary mask.

  Returns:
    numpy.ndarray: A float mask with datatype `equimage.params.imagetype` and values 1 where mask
    is True and 0 where mask is False. If already a float array, the input mask is returned as is.
  """
  return np.asarray(mask, dtype = params.imagetype)

def extend_bmask(mask, extend):
  """Extend or erode a binary mask.

  Args:
    mask (bool numpy.ndarray): The input binary mask.
    extend (int): The number of pixels by which the mask is extended.
      The mask is extended if extend > 0, and eroded if extend < 0.

  Returns:
    numpy.ndarray: The extended binary mask.
  """
  if extend > 0:
    return skimo.isotropic_dilation(mask, extend)
  else:
    return skimo.isotropic_erosion(mask, -extend)

def smooth_mask(mask, radius, mode = "zero"):
  """Smooth a binary or float mask.

  The input mask is converted into a float mask and convolved with a disk of radius smooth.

  Args:
    mask (numpy.ndarray): The input binary or float mask.
    radius (float): The smoothing radius (pixels). The edges of the output float mask get smoothed
      over 2*radius pixels.
    mode (str, optional): How to extend the mask across its boundaries for the convolution:

      - "reflect": the mask is reflected about the edge of the last pixel (abcd → dcba|abcd|dcba).
      - "mirror": the mask is reflected about the center of the last pixel (abcd → dcb|abcd|cba).
      - "nearest": the mask is padded with the value of the last pixel (abcd → aaaa|abcd|dddd).
      - "zero" (default): the mask is padded with zeros (abcd → 0000|abcd|0000).
      - "one": the mask is padded with ones (abcd → 1111|abcd|1111).

  Returns:
    numpy.ndarray: The smoothed, float mask.
  """
  # Translate modes.
  if mode == "zero":
    mode = "constant"
    cval = 0.
  elif mode == "one":
    mode = "constant"
    cval = 1.
  # Convert into a float mask.
  fmask = float_mask(mask)
  # Smooth the float mask.
  if radius > 0.:
    kernel = skimo.disk(radius, dtype = params.imagetype)
    kernel /= np.sum(kernel)
    fmask = ndimg.convolve(fmask, kernel, mode = mode, cval = cval)
  return fmask

def threshold_bmask(filtered, threshold, extend = 0):
  """Set-up a threshold binary mask.

  Returns the pixels of the image such that filtered >= threshold as a binary mask.

  See also:
    :meth:`Image.filter() <.filter>`,
    :func:`threshold_fmask`

  Args:
    filtered (numpy.ndarray): The output of a filter (e.g., local average, ...) applied to the image
      (see :meth:`Image.filter() <.filter>`).
    threshold (float): The threshold for the mask. The mask is True wherever filtered >= threshold,
      and False elsewhere.
    extend (int, optional): Once computed, the mask can be extended/eroded by extend pixels (default 0).
      The mask is is extended if extend > 0, and eroded if extend < 0.

  Returns:
    numpy.ndarray: The mask as a boolean array with the same shape as filtered.
  """
  return extend_bmask(filtered >= threshold, extend)

def threshold_fmask(filtered, threshold, extend = 0, smooth = 0., mode = "zero"):
  """Set-up a threshold float mask.

  Returns the pixels of the image such that filtered >= threshold as a float mask.

  See also:
    :meth:`Image.filter() <.filter>`,
    :func:`smooth_mask`,
    :func:`threshold_bmask`

  Args:
    filtered (numpy.ndarray): The output of a filter (e.g., local average, ...) applied to the image
      (see :meth:`Image.filter() <.filter>`).
    threshold (float): The threshold for the mask. The mask is 1 wherever filtered >= threshold,
      and 0 elsewhere.
    extend (int, optional): Once computed, the mask can be extended/eroded by extend pixels (default 0).
      The mask is is extended if extend > 0, and eroded if extend < 0.
    smooth (float, optional): Once extended, the edges of the mask can be smoothed over 2*smooth pixels
      (default 0; see :func:`smooth_mask`).
    mode (str, optional): How to extend the mask across its boundaries for smoothing:

      - "reflect": the mask is reflected about the edge of the last pixel (abcd → dcba|abcd|dcba).
      - "mirror": the mask is reflected about the center of the last pixel (abcd → dcb|abcd|cba).
      - "nearest": the mask is padded with the value of the last pixel (abcd → aaaa|abcd|dddd).
      - "zero" (default): the mask is padded with zeros (abcd → 0000|abcd|0000).
      - "one": the mask is padded with ones (abcd → 1111|abcd|1111).

  Returns:
    numpy.ndarray: The mask as a float array with the same shape as filtered.
  """
  return smooth_mask(threshold_bmask(filtered, threshold, extend), smooth, mode = mode)

def shape_bmask(shape, x, y, width, height):
    """Return a binary mask defined by the input shape.

    Args:
      shape (str): Either "rectangle" for a rectangle, "ellipse" for an ellipse, or "polygon" for
        a polygon.
      x (tuple, list or numpy.ndarray) : The x coordinates of the shape (pixels along the width).
      y (tuple, list or numpy.ndarray) : The y coordinates of the shape (pixels along the height):

        - If shape = "rectangle", x = (x1, x2) and y = (y1, y2) define the coordinates of two opposite
          corners C1 = (x1, y1) and C2 = (x2, y2) of the rectangle.
        - If shape = "ellipse", x = (x1, x2) and y = (y1, y2) define the coordinates of two opposite
          corners C1 = (x1, y1) and C2 = (x2, y2) of the rectangle that bounds the ellipse.
        - If shape = "polygon", the points P[n] = (x[n], y[n]) (0 <= n < len(x)) are the vertices of
          the polygon.

      width (int): The width of the mask (pixels).
      height (int): The height of the mask (pixels).

    Returns:
      numpy.ndarray: A boolean array with shape (height, width) and values True in the shape and
      False outside.
    """
    if shape == "rectangle":
      if len(x) != 2 or len(y) != 2: raise ValueError("Error, x and y must have exactly two elements for shape = 'rect'.")
      x1 = max(int(np.rint(min(x)))  , 0)
      x2 = min(int(np.rint(max(x)))+1, width)
      y1 = max(int(np.rint(min(y)))  , 0)
      y2 = min(int(np.rint(max(y)))+1, height)
      bmask = np.zeros((height, width), dtype = bool)
      bmask[y1:y2, x1:x2] = True
    elif shape == "ellipse":
      if len(x) != 2 or len(y) != 2: raise ValueError("Error, x and y must have exactly two elements for shape = 'ellipse'.")
      xc = (x[0]+x[1])/2.
      yc = (y[0]+y[1])/2.
      rx = abs(x[1]-x[0])/2.
      ry = abs(y[1]-y[0])/2.
      bmask = np.zeros((height, width), dtype = bool)
      rs, cs = skdraw.ellipse(yc, xc, ry, rx, shape = (height, width))
      bmask[rs, cs] = True
    elif shape == "polygon":
      if len(x) != len(y): raise ValueError("Error, x and y must have the same length for shape = 'polygon'.")
      bmask = skdraw.polygon2mask((height, width), np.column_stack((y, x)))
    else:
      raise ValueError(f"Error, unknown shape '{shape}'.")
    return bmask

#####################################
# For inclusion in the Image class. #
#####################################

class MixinImage:
  """To be included in the Image class."""

  def filter(self, channel, filter, radius, mode = "reflect"):
    """Apply a spatial filter to a selected channel of the image.

    The main purpose of this method is to prepare masks for image processing.

    See also:
      :func:`threshold_bmask`,
      :func:`threshold_fmask`

    Args:
      channel (str): The selected channel:

        - "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          The first/second/third channel (all images).
        - "H": The HSV/HSL hue (RGB, HSV and HSL images).
        - "V": The HSV value (RGB, HSV and grayscale images).
        - "S": The HSV saturation (RGB, HSV and grayscale images).
        - "L'": The HSL lightness (RGB, HSL and grayscale images).
        - "S'": The HSL saturation (RGB, HSL and grayscale images).
        - "L": The luma (RGB and grayscale images).
        - "L*": The CIE lightness L* (RGB, grayscale, CIELab and CIELuv images).
        - "c*": The CIE chroma c* (CIELab and CIELuv images).
        - "s*": The CIE saturation s* (CIELuv images).

      filter (str): The filter:

        - "mean": Return the average of the channel within a disk around each pixel.
        - "median": Return the median of the channel within a disk around each pixel.
        - "gaussian": Return the gaussian average of the channel around each pixel.
        - "maximum": Return the maximum of the channel within a disk around each pixel.

      radius (float): The radius of the disk (pixels). The standard deviation for gaussian average
        is radius/3.
      mode (str, optional): How to extend the image across its boundaries:

        - "reflect" (default): the image is reflected about the edge of the last pixel (abcd → dcba|abcd|dcba).
        - "mirror": the image is reflected about the center of the last pixel (abcd → dcb|abcd|cba).
        - "nearest": the image is padded with the value of the last pixel (abcd → aaaa|abcd|dddd).
        - "zero": the image is padded with zeros (abcd → 0000|abcd|0000).

    Returns:
      numpy.ndarray: The output of the filter as an array with shape (image height, image width).
    """
    if mode == "zero": mode = "constant" # Translate modes.
    data = self.get_channel(channel)
    if filter == "gaussian":
      return ndimg.gaussian_filter(data, sigma = radius/3., mode = mode, cval = 0.)
    elif filter == "mean":
      kernel = skimo.disk(radius, dtype = self.dtype)
      kernel /= np.sum(kernel)
      return ndimg.convolve(data, kernel, mode = mode, cval = 0.)
    elif filter == "median":
      return ndimg.median_filter(data, footprint = skimo.disk(radius, dtype = bool), mode = mode, cval = 0.)
    elif filter == "maximum":
      return ndimg.maximum_filter(data, footprint = skimo.disk(radius, dtype = bool), mode = mode, cval = 0.)
    else:
      raise ValueError(f"Error, unknown filter '{filter}'.")

  def shape_bmask(self, shape, x, y):
    """Return a binary mask defined by the input shape.

    Args:
      shape (str): Either "rectangle" for a rectangle, "ellipse" for an ellipse, or "polygon" for
        a polygon.
      x (tuple, list or numpy.ndarray) : The x coordinates of the shape (pixels along the width).
      y (tuple, list or numpy.ndarray) : The y coordinates of the shape (pixels along the height):

        - If shape = "rectangle", x = (x1, x2) and y = (y1, y2) define the coordinates of two opposite
          corners C1 = (x1, y1) and C2 = (x2, y2) of the rectangle.
        - If shape = "ellipse", x = (x1, x2) and y = (y1, y2) define the coordinates of two opposite
          corners C1 = (x1, y1) and C2 = (x2, y2) of the rectangle that bounds the ellipse.
        - If shape = "polygon", the points P[n] = (x[n], y[n]) (0 <= n < len(x)) are the vertices of
          the polygon.

    Returns:
      numpy.ndarray: A boolean array with shape (image height, image width), and values True in the
      shape and False outside.
    """
    width, height = self.get_size()
    return shape_bmask(shape, x, y, width, height)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""eQuimage top-level symbols.

This imports relevant symbols from the submodules into the equimage/equimagelab namespace.
These symbols are defined by the :py:class:`__all__` dictionary (if any) of each submodule, and
listed in their docstring.

Also, the methods of the :py:class:`MixinImage` class of each submodule are imported in the Image class.
"""

from .params import *
from .image import *
from .image_utils import *
from .image_colorspaces import *
from .image_colors import *
from .image_stretch import *
from .image_masks import *
from .image_multiscale import *
from .image_io import *



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""eQuimageLab package."""

__version__ = "1.4.1"
__packagepath__ = __path__[0]

# Import top-level symbols.

from equimage.imports import *
from .jupyter.imports import *
from . import params

print("######################################"+len(__version__)*"#")
print(f"# Welcome to eQuimageLab version {__version__}... #")
print("######################################"+len(__version__)*"#")
print("Documentation available at: https://astro.ymniquet.fr/codes/equimagelab/docs/")

set_RGB_luma("human")



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Image class.

The following symbols are imported in the equimage/equimagelab namespaces for convenience:
  "Image".
"""

__all__ = ["Image"]

import numpy as np

from . import params
from . import image_colorspaces
from . import image_utils
from . import image_geometry
from . import image_colors
from . import image_stretch
from . import image_filters
from . import image_skimage
from . import image_multiscale
from . import image_masks
from . import image_stats
from . import image_editors
from . import image_stars
from . import image_io

class Image(np.lib.mixins.NDArrayOperatorsMixin,
            image_colorspaces.MixinImage, image_utils.MixinImage, image_geometry.MixinImage,
            image_colors.MixinImage, image_stretch.MixinImage, image_filters.MixinImage, image_skimage.MixinImage,
            image_multiscale.MixinImage, image_masks.MixinImage, image_stats.MixinImage, image_editors.MixinImage,
            image_stars.MixinImage, image_io.MixinImage):
  """Image class.

  The image is stored as self.image, a numpy.ndarray with dtype numpy.float32 or numpy.float64.
  Color images are represented as arrays with shape (3, height, width) and grayscale images as
  arrays with shape (1, height, width). The leading axis spans the color channels, and the last
  two the height and width of the image.

  The class embeds colorspace and colormodel attributes for the color space and model of the image.

  The colorspace attribute can be:

    - "lRGB" for the linear RGB color space.
    - "sRGB" for the sRGB color space.
    - "CIELab" for the CIELab color space.
    - "CIELuv" for the CIELuv color space.

  In the lRGB and sRGB color spaces, the colormodel attribute can be:

    - "gray": grayscale image with one single channel within [0, 1].
    - "RGB": the 3 channels of the image are the red, blue, and green levels within [0, 1].
    - "HSV": the 3 channels of the image are the HSV hue, saturation and value within [0, 1].
    - "HSL": the 3 channels of the image are the HSL hue, saturation and lightness within [0, 1].

  In the CIELab color space, the colormodel attribute can be:

    - "Lab": the 3 channels of the image are the CIELab components L*/100, a*/100 and b*/100.
      The lightness L*/100 fits within [0, 1], but a* and b* are signed and not bounded.
    - "Lch": the 3 channels of the image are the CIELab components L*/100, c*/100 and h*/(2π).
      The lightness L*/100 and the reduced hue angle h*/(2π) fit within [0, 1], but the
      chroma c* is not bounded by 1.

  In the CIELuv color space, the colormodel attribute can be:

    - "Luv": the 3 channels of the image are the CIELuv components L*/100, u*/100 and v*/100.
      The lightness L*/100 fits within [0, 1], but u* and v* are signed and not bounded.
    - "Lch": the 3 channels of the image are the CIELuv components L*/100, c*/100 and h*/(2π).
      The lightness L*/100 and the reduced hue angle h*/(2π) fit within [0, 1], but the
      chroma c* is not bounded by 1.
    - "Lsh": the 3 channels of the image are the CIELuv components L*/100, s*/100 and h*/(2π).
      The lightness L*/100 and the reduced hue angle h*/(2π) fit within [0, 1], but the
      saturation s* = c*/L* is not bounded by 1.

  The default color space is sRGB and the default color model is RGB.

  The dtype of the images (numpy.float32 or numpy.float64) can be set with :meth:`params.set_image_type() <equimage.params.set_image_type>`.
  """

  ################
  # Constructor. #
  ################

  def __init__(self, image, channels = 0, colorspace = "sRGB", colormodel = "RGB"):
    """Initialize a new Image object with the input image.

    Args:
      image (numpy.ndarray or Image): The input image.
      channels (int, optional): The position of the channel axis for color images (default 0).
      colorspace (str, optional): The color space of the image (default "sRGB").
        Can be "lRGB" (linear RGB color space), "sRGB" (sRGB color space), "CIELab" (CIELab colorspace),
        or "CIELuv" (CIELuv color space).
      colormodel (str, optional): The color model of the image (default "RGB").
        In the lRGB/SRGB color spaces, can be "RGB" (RGB color model), "HSV" (HSV color model), "HSL"
        (HSL color model) or "gray" (grayscale image).
        In the CIELab color space, can be "Lab" (L*a*b* color model) or "Lch" (L*c*h* color model).
        In the CIELuv color space, can be "Luv" (L*u*v* color model), "Lch" (L*c*h* color model)
        or "Lsh" (L*s*h* model).
    """
    # Check color space and model.
    if colorspace in ["lRGB", "sRGB"]:
      if colormodel not in ["RGB", "HSV", "HSL", "gray"]:
        raise ValueError(f"Error, the color model of {colorspace} images must be 'RGB', 'HSV', 'HSL' or 'gray' (got '{colormodel}').")
    elif colorspace == "CIELab":
      if colormodel not in ["Lab", "Lch"]:
        raise ValueError(f"Error, the color model of {colorspace} images must be 'Lab' or 'Lch' (got '{colormodel}').")
    elif colorspace == "CIELuv":
      if colormodel not in ["Luv", "Lch", "Lsh"]:
        raise ValueError(f"Error, the color model of {colorspace} images must be 'Luv', 'Lch' or 'Lsh' (got '{colormodel}').")
    else:
      raise ValueError(f"Error, the color space must be 'lRGB', 'sRGB', 'CIELab' or 'CIELuv' (got '{colorspace}').")
    # Convert the input image into an array.
    image = np.asarray(image, dtype = params.imagetype)
    # Validate the image.
    if image.ndim == 2:
      colormodel = "gray"  # Enforce colormodel = "gray".
      image = np.expand_dims(image, axis = 0)
    elif image.ndim == 3:
      if channels != 0: image = np.moveaxis(image, channels, 0)
      nc = image.shape[0]
      if nc == 1:
        colormodel = "gray" # Enforce colormodel = "gray".
      elif nc == 3:
        if colormodel == "gray":
          raise ValueError(f"Error, a grayscale image must have one single channel (found {nc}).")
      else:
        raise ValueError(f"Error, an image must have either 1 or 3 channels (found {nc}).")
    else:
      raise ValueError(f"Error, an image must have either 2 or 3 dimensions (found {image.ndim}).")
    if colorspace in ["CIELab", "CIELuv"] and colormodel == "gray":
      raise ValueError(f"Error, a CIELab/CIELuv image must have 3 channels.")
    # Register image, color space and model.
    self.image = image
    self.dtype = self.image.dtype # Add a reference to image type.
    self.colorspace = colorspace
    self.colormodel = colormodel

  def newImage(self, image, **kwargs):
    """Return a new Image object with the input image (with, by default, the same color space and
    color model as self).

    Args:
      image (numpy.ndarray): The input image.
      colorspace (str, optional): The color space of the image (default self.colorspace).
        Can be "lRGB" (linear RGB color space), "sRGB" (sRGB color space), or "CIELab" (CIELab color space).
      colormodel (str, optional): The color model of the image (default self.colormodel).
        In the lRGB/SRGB color spaces, can be "RGB" (RGB color model), "HSV" (HSV color model), "HSL"
        (HSL color model) or "gray" (grayscale image).
        In the CIELab color space, can be "Lab" (L*a*b* color model) or "Lch" (L*c*h* color model).
        In the CIELuv color space, can be "Luv" (L*u*v* color model), "Lch" (L*c*h* color model)
        or "Lsh" (L*s*h* model).

    Returns:
      Image: The new Image object.
    """
    colorspace = kwargs.pop("colorspace", self.colorspace)
    colormodel = kwargs.pop("colormodel", self.colormodel)
    if kwargs: print("Discarding extra keyword arguments in Image.newImage...")
    return Image(image, colorspace = colorspace, colormodel = colormodel)

  def copy(self):
    """Return a copy of the object.

    Returns:
      Image: A copy of the object.
    """
    return self.newImage(self.image.copy())

  ######################
  # Object management. #
  ######################

  def __repr__(self):
    """Return the object representation."""
    return f"{self.__class__.__name__}(size = {self.image.shape[2]}x{self.image.shape[1]} pixels, colorspace = {self.colorspace}, colormodel = {self.colormodel}, type = {self.dtype})"

  def __getitem__(self, channels):
    """Return channel(s) of the image.

    Implements the indexer operator self[channels] as self.image[channels].

    Args:
      channels (int or slice): The channels to be returned.

    Returns
      numpy.ndarray: The channels self.image[channels].
    """
    return self.image[channels]

  def __setitem__(self, channels, data):
    """Set channel(s) of the image.

    Implements the operation self.image[channels] = data as self[channels] = data.

    Args:
      channels (int or slice): The channels to be set.
      data (numpy.ndarray): The new channels data.
    """
    self.image[channels] = data

  def __array__(self, dtype = None, copy = None):
    """Expose an Image object as an numpy.ndarray."""
    return np.array(self.image, dtype = dtype, copy = copy)

  def __array_ufunc__(self, ufunc, method, *args, **kwargs):
    """Apply numpy ufuncs to an Image object."""
    if method != "__call__": raise NotImplemented(f"Operation {method} not implemented on Image ufuncs.")
    reference = None
    mixed = False
    inputs = []
    for arg in args:
      if isinstance(arg, Image):
        if reference is None:
          reference = arg
        else:
          if arg.colorspace != reference.colorspace or arg.colormodel != reference.colormodel and not mixed:
            print("Warning ! This operation mixes images with different color spaces or models !..")
            mixed = True
        inputs.append(arg.image)
      else:
        inputs.append(arg)
    if (out := kwargs.get("out", None)) is not None: # In-place operation.
      outputs = []
      for arg in out:
        if isinstance(arg, Image):
          if arg.colorspace != reference.colorspace or arg.colormodel != reference.colormodel and not mixed:
            print("Warning ! This operation mixes images with different color spaces or models !..")
            mixed = True
          outputs.append(arg.image)
        else:
          outputs.append(arg)
      kwargs["out"] = tuple(outputs)
    output = ufunc(*inputs, **kwargs)
    if isinstance(output, np.ndarray): # Note: If there are multiple outputs, we don't convert these to Images.
      if output.ndim == 0: # Is output actually a scalar ?
        return output[()]
      else:
        if not mixed and output.shape == reference.image.shape:
          return Image(output, colorspace = reference.colorspace, colormodel = reference.colormodel)
        else:
          return output
    else:
      return output

  def __array_function__(self, func, types, args, kwargs):
    """Apply numpy array functions to an Image object."""
    reference = None
    mixed = False
    inputs = []
    for arg in args:
      if isinstance(arg, Image):
        if reference is None:
          reference = arg
        else:
          if arg.colorspace != reference.colorspace or arg.colormodel != reference.colormodel and not mixed:
            print("Warning ! This operation mixes images with different color spaces or models !..")
            mixed = True
        inputs.append(arg.image)
      else:
        inputs.append(arg)
    output = func(*inputs, **kwargs)
    if isinstance(output, np.ndarray):
      if output.ndim == 0: # Is output actually a scalar ?
        return output[()]
      else:
        if not mixed and output.shape == reference.image.shape:
          return Image(output, colorspace = reference.colorspace, colormodel = reference.colormodel)
        else:
          return output
    else:
      return output

  ##################
  # Image queries. #
  ##################

  def get_image(self, channels = 0, copy = False):
    """Return the image data.

    Args:
      channels (int, optional): The position of the channel axis (default 0).
      copy (bool, optional): If True, return a copy of the image data;
                             If False (default), return a view.

    Returns:
      numpy.ndarray: The image data.
    """
    image = self.image
    if channels != 0: image = np.moveaxis(image, 0, channels)
    return image.copy() if copy else image

  def get_shape(self):
    """Return the shape of the image data.

    Returns:
      tuple: (number of channels, height of the image in pixels, width of the image in pixels).
    """
    return self.image.shape

  def get_size(self):
    """Return the width and height of the image.

    Returns:
      tuple: (width, height) of the image in pixels.
    """
    return self.image.shape[2], self.image.shape[1]

  def get_nc(self):
    """Return the number of channels of the image.

    Returns:
      int: The number of channels of the image.
    """
    return self.image.shape[0]

  def get_color_space(self):
    """Return the color space of the image.

    Returns:
      str: The color space of the image.
    """
    return self.colorspace

  def get_color_model(self):
    """Return the color model of the image.

    Returns:
      str: The color model of the image.
    """
    return self.colormodel

  ######################
  # Image conversions. #
  ######################

  def int8(self):
    """Return the image as an array of 8 bits integers with shape (height, width, channels).

    Warning:
      This method maps [0., 1.] onto [0, 255].
      Not suitable for the CIELab and CIELuv color spaces !

    Returns:
      numpy.ndarray: The image as an array of 8 bits integers with shape (height, width, channels).
    """
    image = self.get_image(channels = -1)
    data = np.clip(image*255, 0, 255)
    return np.rint(data).astype("uint8")

  def int16(self):
    """Return the image as an array of 16 bits integers with shape (height, width, channels).

    Warning:
      This method maps [0., 1.] onto [0, 65535].
      Not suitable for the CIELab and CIELuv color spaces !

    Returns:
      numpy.ndarray: The image as an array of 16 bits integers with shape (height, width, channels).
    """
    image = self.get_image(channels = -1)
    data = np.clip(image*65535, 0, 65535)
    return np.rint(data).astype("uint16")

  def int32(self):
    """Return the image as an array of 32 bits integers with shape (height, width, channels).

    Warning:
      This method maps [0., 1.] onto [0, 4294967295].
      Not suitable for the CIELab and CIELuv color spaces !

    Returns:
      numpy.ndarray: The image as an array of 32 bits integers with shape (height, width, channels).
    """
    image = self.get_image(channels = -1)
    data = np.clip(image*4294967295, 0, 4294967295)
    return np.rint(data).astype("uint32")



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""External image editors."""

import os
import tempfile
import subprocess
import numpy as np

from . import image_io as io

#####################################
# For inclusion in the Image class. #
#####################################

class MixinImage:
  """To be included in the Image class."""

  def edit_with(self, command, export = "tiff", depth = 16, script = None, editor = "<Editor>", interactive = True, cwd = None):
    """Edit the image with an external tool.

    The image is saved on disk; the editor is then run on this file, which is finally reloaded in
    eQuimageLab and returned.

    The user/editor must simply overwrite the edited file on exit.

    Args:
      command (str): The editor command to be run (e.g., "gimp -n $FILE$").
        Any "$FILE$" is replaced by the name of the image file to be opened by the editor.
        Any "$SCRIPT$" is replaced by the name of the script file to be run by the editor.
      export (str, optional): The format used to export the image. Can be:

        - "png": PNG file with depth = 8 or 16 bits integer per channel.
        - "tiff" (default): TIFF file with depth = 8, 16 or 32 bits integer per channel.
        - "fits": FITS file width depth = 32 bits float per channel.

      depth (int, optional): The color depth (bits per channel) used to export the image
        (see above; default 16).
      script (str, optional): A string containing a script to be executed by the editor
        (default None). Any "$FILE$" is replaced by the name of the image file.
      editor (str, optional): The name of the editor (for pretty print; default "<Editor>").
      interactive (bool, optional): If True (default), the editor is interactive (awaits commands
        from the user); if False, the editor processes the image autonomously and does not require
        inputs from the user.
      cwd (str, optional): If not None (default), change to this working directory before running
        the editor.

    Returns:
      Image: The edited image.
    """
    self.check_color_model("RGB", "gray")
    if export not in ["png", "tiff", "fits"]:
      raise ValueError(f"Error, unknown export format '{export}'.")
    if depth not in [8, 16, 32]:
      raise ValueError("Error, depth must be 8, 16 or 32 bpc.")
    # Edit image.
    with tempfile.TemporaryDirectory(ignore_cleanup_errors = True) as tmpdir:
      filename = "eQuimageLab."+export # Set image tmp file name.
      filepath = os.path.join(tmpdir, filename)
      filefound = False
      # Process script.
      if script is not None:
        scriptname = "eQuimageLab.script" # Set script tmp file name.
        scriptpath = os.path.join(tmpdir, scriptname)
        scriptfound = False
        filefound = "$FILE$" in script
        if filefound: script = script.replace("$FILE$", filepath)
      # Process command.
      splitcmd = []
      for item in command.strip().split(" "):
        if item == "$FILE$":
          filefound = True
          splitcmd.append(filepath)
        elif item == "$SCRIPT$":
          if script is None:
            raise ValueError("Error, there is a place holder for a script file ($SCRIPT$) in the command but no script provided.")
          scriptfound = True
          splitcmd.append(scriptpath)
        else:
          splitcmd.append(item)
      # Check script & command consistency.
      if script is None:
        if not filefound:
          raise ValueError("Error, no place holder for the image file ($FILE$) found in the command.")
      else:
        if not filefound:
          raise ValueError("Error, no place holder for the image file ($FILE$) found in the command and in the script.")
        if not scriptfound:
          raise ValueError("Error, no place holder for the script file ($SCRIPT$) found in the command.")
      # Save script.
      if script is not None:
        print(f"Writing file {scriptpath}...")
        with open(scriptpath, "w") as f: f.write(script)
      # Save image.
      print(f"Writing file {filepath} with depth = {depth} bpc...")
      self.save(filepath, depth = depth, compress = 0, verbose = False) # Don't compress image to ensure compatibility with the editor.
      ctime = os.path.getmtime(filepath)
      # Run editor.
      print(f"Running {editor}...")
      if interactive: print(f"Overwrite file {filepath} when leaving {editor}.")
      subprocess.run(splitcmd, cwd = cwd)
      # Load and return edited image.
      mtime = os.path.getmtime(filepath)
      if mtime == ctime:
        if interactive:
          print(f"The image has not been modified by {editor}; Returning the original...")
          return self.copy()
        else:
          raise ValueError(f"Error, the image has not been modified by {editor}...")
      print(f"Reading file {filepath}...")
      image, meta = io.load_image_as_array(filepath, verbose = False)
      return self.newImage(image)

  def edit_with_gimp(self, export = "tiff", depth = 16):
    """Edit the image with Gimp.

    The image is saved on disk; Gimp is then run on this file, which is finally reloaded in
    eQuimageLab and returned.

    The user must simply overwrite the edited file when leaving Gimp.

    Warning:
      The command "gimp" must be in the PATH.

    Args:
      export (str, optional): The format used to export the image. Can be:

        - "png": PNG file with depth = 8 or 16 bits integer per channel.
        - "tiff" (default): TIFF file with depth = 8, 16 or 32 bits integer per channel.
        - "fits": FITS file width depth = 32 bits float per channel.

      depth (int, optional): The color depth (bits per channel) used to export the image
        (see above; default 16).

    Returns:
      Image: The edited image.
    """
    return self.edit_with("gimp -n $FILE$", export = export, depth = depth, editor = "Gimp", interactive = True)

  def edit_with_siril(self):
    """Edit the image with Siril.

    The image is saved as a FITS file (32 bits float per channel); Siril is then run on this file,
    which is finally reloaded in eQuimageLab and returned.

    The user must simply overwrite the edited file when leaving Siril.

    Warning:
      The command "siril" must be in the PATH.

    Returns:
      Image: The edited image.
    """
    return self.edit_with("siril $FILE$", export = "fits", depth = 32, editor = "Siril", interactive = True)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Plotly backend for JupyterLab interface.

The following symbols are imported in the equimagelab namespace for convenience:
  "show", "show_t", "show_histograms", "show_statistics", "light_curve".
"""

__all__ = ["show", "show_t", "show_histograms", "show_statistics", "light_curve"]

import numpy as np
import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
from plotly.subplots import make_subplots
pio.renderers.default = "jupyterlab"

from equimage import Image
from equimage.image_stats import parse_channels

from . import params
from .utils import format_images

#####################
# Helper functions. #
#####################

def _figure_formatted_image_(image, dx = 1, dy = 1, width = -1, hover = False, template = "plotly_dark"):
  """Prepare a ploty figure for the input (formatted) image.

  Args:
    image (numpy.ndarray): The prepared formatted (namely, processed by :meth:`jupyter.utils.format_images() <.format_images>`).
    dx (int, optional): The size of a pixel along x (default 1).
    dy (int, optional): The size of a pixel along y (default 1).
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    hover (bool, optional): If True, show the image data on hover (default False).
      Warning: setting hover = True can slow down display a lot !
    template (str, optional): The template for the figure (default "plotly_dark").

  Returns:
    plotly.graph_objects.Figure: A plotly figure with the image.
  """
  if width <= 0: width = params.maxwidth
  # Plot image.
  x = np.arange(0, image.shape[1])*dx
  y = np.arange(0, image.shape[0])*dy
  figure = px.imshow(image, x = x, y = y, zmin = 0., zmax = 1., aspect = "equal", binary_string = not hover)
  figure.update_traces(name = "", hovertemplate = "(%{x}, %{y}): %{z}" if hover else "(%{x}, %{y})")
  layout = go.Layout(template = template,
                     width = width+params.lmargin+params.rmargin, height = width*image.shape[0]/image.shape[1]+params.bmargin+params.tmargin,
                     margin = go.layout.Margin(l = params.lmargin, r = params.rmargin, b = params.bmargin, t = params.tmargin, autoexpand = True))
  figure.update_layout(layout)
  return figure

def _figure_image_(image, sampling = -1, width = -1, hover = False, template = "plotly_dark"):
  """Prepare a ploty figure for the input image.

  Args:
    image: An Image object or numpy.ndarray with shape (height, width, 3) (for a color image),
      (height, width, 1) or (height, width) (for a grayscale image).
    sampling (int, optional): The downsampling rate (defaults to `jupyter.params.sampling` if negative).
      Only image[::sampling, ::sampling] is shown, to speed up display.
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    hover (bool, optional): If True, show the image data on hover (default False).
      Warning: setting hover = True can slow down display a lot !
    template (str, optional): The template for the figure (default "plotly_dark").

  Returns:
    plotly.graph_objects.Figure: A plotly figure with the image.
  """
  if sampling <= 0: sampling = params.sampling
  return _figure_formatted_image_(format_images(image, sampling = sampling), dx = sampling, dy = sampling, width = width, hover = hover, template = template)

def _figure_histograms_(image, channels = "", log = True, width = -1, xlabel = "Level", trans = None, template = "plotly_dark"):
  """Prepare a plotly figure with the histograms of an image.

  Args:
    image: An Image object or numpy.ndarray with shape (height, width, 3) (for a color image),
      (height, width, 1) or (height, width) (for a grayscale image).
    channels (str, optional): The channels of the histograms (default "" = "RGBL" for red, green,
      blue, luma).
    log (bool, optional): If True (default), plot the histogram counts in log scale.
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    xlabel (str, optional): The x axis label of the plot (default "Level").
    trans (optional): A container with an histogram transformation (see :meth:`Image.apply_channels() <.apply_channels>`),
      plotted on top of the histograms (default None).
    template (str, optional): The template for the figure (default "plotly_dark").

  Returns:
    plotly.graph_objects.Figure: A plotly figure with the histograms of the image.
  """
  if width <= 0: width = params.maxwidth
  # Prepare histograms.
  if not issubclass(type(image), Image): image = Image(image, channels = -1)
  if channels == "":
    hists = getattr(image, "hists", None)
    if hists is None: hists = image.histograms()
  else:
    hists = image.histograms(channels = channels)
  # Set-up lines.
  mline = dict(color = params.mlinecolor)
  mline2 = dict(color = params.mlinecolor, width = 2)
  mlinedot1 = dict(color = params.mlinecolor, dash = "dot", width = 1)
  mlinedash1 = dict(color = params.mlinecolor, dash = "dash", width = 1)
  mlinedashdot1 = dict(color = params.mlinecolor, dash = "dashdot", width = 1)
  # Plot histograms.
  figure = make_subplots(specs = [[dict(secondary_y = trans is not None, r = -0.06)]])
  updatemenus = []
  n = len(hists)
  for channel in hists.values():
    midpoints = (channel.edges[1:]+channel.edges[:-1])/2.
    figure.add_trace(go.Scatter(x = midpoints, y = channel.counts, name = channel.name, mode = "lines", line = dict(color = channel.color, width = 2)),
                     secondary_y = False)
  figure.add_vline(x = 0., line = mlinedashdot1, secondary_y = False)
  figure.add_vline(x = 1., line = mlinedashdot1, secondary_y = False)
  figure.update_xaxes(title_text = xlabel, ticks = "inside", rangemode = "tozero")
  figure.update_yaxes(title_text = "Count", ticks = "inside", rangemode = "tozero", secondary_y = False)
  if log:
    figure.update_yaxes(type = "log", secondary_y = False)
    active =  0
  else:
    active = -1
  # Add lin/log toggle button.
  ybutton = 1.+.025*1024./width
  buttons = [dict(label = "lin/log", method = "relayout", args = [{"yaxis.type": "log"}], args2 = [{"yaxis.type": "linear"}])]
  updatemenus.append(dict(type = "buttons", buttons = buttons, active = active, showactive = False,
                          xanchor = "left", x = 0., yanchor = "bottom", y = ybutton))
  # Plot transformation.
  if trans is not None:
    if trans.type == "hist":
      cef = np.log(np.maximum(np.gradient(trans.y, trans.x), 1.e-12))
      figure.add_trace(go.Scatter(x = trans.x, y = trans.y, name = trans.ylabel, mode = "lines", line = mline2, showlegend = False),
                       secondary_y = True)
      m = 2 if hasattr(trans, "xm") else 1
      if m == 2:
        figure.add_trace(go.Scatter(x = trans.xm, y = trans.ym, mode = "markers", marker = dict(size = 8, color = params.mlinecolor), showlegend = False),
                         secondary_y = True)
      figure.add_trace(go.Scatter(x = trans.x, y = cef, name = f"log {trans.ylabel}'", mode = "lines", line = mline2, showlegend = False, visible = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = [0., 1.], y = [0., 0.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = [0., 1.], y = [1., 1.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = [0., 1.], y = [0., 1.], name = "", mode = "lines", line = mlinedot1, showlegend = False),
                       secondary_y = True)
      if hasattr(trans, "xticks"):
        for xtick in trans.xticks:
          figure.add_vline(x = xtick, line = mlinedash1, secondary_y = True)
      ftitle = f"{trans.ylabel}({trans.channels})"
      ceftitle = f"log {trans.ylabel}'({trans.channels})"
      figure.update_yaxes(title_text = ftitle, title_font = mline, ticks = "inside", tickfont = mline, showgrid = False, rangemode = "tozero", secondary_y = True)
      # Add f/log f' toggle button.
      buttons = [dict(label = "f/log f'", method = "update",
                      args  = [{"visible": n*[True]+m*[False]+[True , True, False, False]}, {"yaxis2.title": ceftitle}],
                      args2 = [{"visible": n*[True]+m*[True ]+[False, True, True , True ]}, {"yaxis2.title": ftitle}])]
      xbutton = .066*1024./width
      updatemenus.append(dict(type = "buttons", buttons = buttons, active = -1, showactive = False,
                              xanchor = "left", x = xbutton, yanchor = "bottom", y = ybutton))
    elif trans.type == "hue":
      figure.add_trace(go.Scatter(x = trans.x, y = trans.y, name = trans.ylabel, mode = "lines", line = mline2, showlegend = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = trans.xm, y = trans.ym, name = trans.ylabel, mode = "markers", marker = dict(size = 16, color = trans.cm), showlegend = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = [0., 1.], y = [0., 0.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False),
                       secondary_y = True)
      figure.update_xaxes(range = [0., 1.])
      figure.update_yaxes(title_text = trans.ylabel, title_font = mline, ticks = "inside", tickfont = mline, showgrid = False, rangemode = "tozero", secondary_y = True)
    else:
      print(f"Can not handle transformations of type '{trans.type}'.")
  # Finalize layout.
  xlegend = 1.+.05*1024./width
  layout = go.Layout(template = template,
                     width = width+params.lmargin+params.rmargin, height = width/3+params.bmargin+params.tmargin,
                     margin = go.layout.Margin(l = params.lmargin, r = params.rmargin, b = params.bmargin, t = params.tmargin, autoexpand = True),
                     legend = dict(xanchor = "left", x = xlegend, yanchor = "top", y = 1.))
  figure.update_layout(layout, updatemenus = updatemenus)
  return figure

def _figure_statistics_(image, channels = "", width = -1, rowheight = -1, template = "plotly_dark"):
  """Prepare a plotly table with the statistics of an image.

  Args:
    image: An Image object or numpy.ndarray with shape (height, width, 3) (for a color image),
      (height, width, 1) or (height, width) (for a grayscale image).
    channels (str, optional): The channels of the statistics (default "" = "RGBL" for red, green,
      blue, luma).
    width (int, optional): The width of the table (defaults to `jupyter.params.maxwidth` if negative).
    rowheight (int, optional): The height of the rows (default to jupyter.params.rowheight if negative).
    template (str, optional): The template for the figure (default "plotly_dark").

  Returns:
    plotly.graph_objects.Figure: A plotly figure with the table of the statistics of the image.
  """
  if width <= 0: width = params.maxwidth
  if rowheight <= 0: rowheight = params.rowheight
  # Prepare statistics.
  if not issubclass(type(image), Image): image = Image(image, channels = -1)
  if channels == "":
    stats = getattr(image, "stats", None)
    if stats is None: stats = image.statistics()
  else:
    stats = image.statistics(channels = channels)
  # Create table.
  columns = [[], [], [], [], [], [], [], []]
  for channel in stats.values():
    columns[0].append(channel.name)
    columns[1].append(f"{channel.minimum:.5f}")
    if channel.percentiles is not None:
      columns[2].append(f"{channel.percentiles[0]:.5f}")
      columns[3].append(f"{channel.percentiles[1]:.5f}")
      columns[4].append(f"{channel.percentiles[2]:.5f}")
    else:
      columns[2].append("None")
      columns[3].append("None")
      columns[4].append("None")
    columns[5].append(f"{channel.maximum:.5f}")
    columns[6].append(f"{channel.zerocount} ({100.*channel.zerocount/channel.npixels:.2f}%)")
    columns[7].append(f"{channel.outcount} ({100.*channel.outcount/channel.npixels:.2f}%)")
  align = ["left"]+7*["right"]
  header = dict(values = ["Channel", "Minimum", "25%", "50%", "75%", "Maximum", "Shadowed", "Highlighted"], align = align, height = rowheight)
  cells = dict(values = columns, align = align, height = rowheight)
  table = go.Table(header = header, cells = cells, columnwidth = [1, 1, 1, 1, 1, 1, 1.5, 1.5])
  # Create figure.
  figure = go.Figure(data = table)
  layout = go.Layout(template = template,
                     width = width+params.lmargin+params.rmargin, height = (len(stats)+1)*rowheight+params.bmargin+params.tmargin,
                     margin = go.layout.Margin(l = params.lmargin, r = params.rmargin, b = 32, t = 32, autoexpand = True))
  figure.update_layout(layout)
  return figure

#####################
# Plotly interface. #
#####################

def show(image, histograms = False, statistics = False, sampling = -1, width = -1, hover = False, renderer = None):
  """Show an image using plotly.

  Args:
    image: An Image object or numpy.ndarray with shape (height, width, 3) (for a color image),
      (height, width, 1) or (height, width) (for a grayscale image).
    histograms (optional): If True or a string, show the histograms of the image. The string lists
      the channels of the histograms (see :meth:`Image.histograms() <.histograms>`). True is
      substituted with "RGBL" (red, green, blue, luma). Default is False.
    statistics (optional): If True or a string, show the statistics of the image. The string lists
      the channels of the statistics (see :meth:`Image.statistics() <.statistics>`). True is
      substituted with "RGBL" (red, green, blue, luma). Default is False.
    sampling (int, optional): The downsampling rate (defaults to `jupyter.params.sampling` if negative).
      Only image[::sampling, ::sampling] is shown, to speed up display.
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    hover (bool, optional): If True, show the image data on hover (default False).
      Warning: setting hover = True can slow down display a lot !
    renderer (str, optional): The plotly renderer (default None = "jupyterlab").
  """
  figure = _figure_image_(image, sampling = sampling, width = width, hover = hover)
  figure.show(renderer)
  if histograms is not False:
    if histograms is True: histograms = ""
    show_histograms(image, channels = histograms, width = width, renderer = renderer)
  if statistics is not False:
    if statistics is True: statistics = ""
    show_statistics(image, channels = statistics, width = width, renderer = renderer)

def show_histograms(image, channels = "", log = True, width = -1, xlabel = "Level", trans = None, renderer = None):
  """Plot the histograms of an image using plotly.

  Args:
    image: An Image object or numpy.ndarray with shape (height, width, 3) (for a color image),
      (height, width, 1) or (height, width) (for a grayscale image).
    channels (str, optional): The channels of the histograms (default "" = "RGBL" for red, green,
      blue, luma). See :meth:`Image.histograms() <.histograms>`.
    log (bool, optional): If True (default), plot the histogram counts in log scale.
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    xlabel (str, optional): The x axis label of the plot (default "Level").
    trans (optional): A container with an histogram transformation (see :meth:`Image.apply_channels() <.apply_channels>`),
      plotted on top of the histograms (default None).
    renderer (str, optional): The plotly renderer (default None = "jupyterlab").
  """
  figure = _figure_histograms_(image, channels = channels, log = log, width = width, xlabel = xlabel, trans = trans)
  if figure is not None: figure.show(renderer)

def show_statistics(image, channels = "", width = -1, rowheight = -1, renderer = None):
  """Display a table with the statistics of an image using plotly.

  Args:
    image: An Image object or numpy.ndarray with shape (height, width, 3) (for a color image),
      (height, width, 1) or (height, width) (for a grayscale image).
    channels (str, optional): The channels of the statistics (default "" = "RGBL" for red, green,
      blue, luma). See :meth:`Image.statistics() <.statistics>`.
    width (int, optional): The width of the table (defaults to `jupyter.params.maxwidth` if negative).
    rowheight (int, optional): The height of the rows (default to `jupyter.params.rowheight` if negative).
    renderer (str, optional): The plotly renderer (default None = "jupyterlab").
  """
  figure = _figure_statistics_(image, channels = channels, width = width, rowheight = rowheight)
  if figure is not None: figure.show(renderer)

def show_t(image, channels = "RGBL", sampling = -1, width = -1, hover = False, renderer = None):
  """Show an image embedding an histogram transformation using plotly.

  Displays the input histograms with the transformation curve, the output histograms, and the output
  image.

  Args:
    image (Image): The output image (must embed a transformation image.trans -
      see :meth:`Image.apply_channels() <.apply_channels>`).
    channels (str, optional): The channels of the histograms (default "" = "RGBL" for red, green,
      blue, luma). The channels of the transformation are automatically appended.
      See :meth:`Image.histograms() <.histograms>`.
    sampling (int, optional): The downsampling rate (defaults to `jupyter.params.sampling` if negative).
      Only image[::sampling, ::sampling] is shown, to speed up display.
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    hover (bool, optional): If True, show the image data on hover (default False).
      Warning: setting hover = True can slow down display a lot !
    renderer (str, optional): The plotly renderer (default None = "jupyterlab").
  """
  if not issubclass(type(image), Image):
    print("The transformations can only be displayed for Image objects.")
    return
  trans = getattr(image, "trans", None)
  if trans is None:
    print("There is no transformation embedded in the input image.")
    return
  reference = trans.input
  keys = parse_channels(channels)
  for key in parse_channels(trans.channels, errors = False):
    if not key in keys: channels += key
  show_histograms(reference, channels = channels, log = True, width = width, xlabel = "Input level", trans = trans, renderer = renderer)
  show_histograms(image, channels = channels, log = True, width = width, xlabel = "Output level", renderer = renderer)
  show(image, histograms = False, statistics = False, sampling = sampling, width = width, hover = hover, renderer = renderer)

def light_curve(image, reference, maxpoints = 32768, width = -1, renderer = None):
  """Plot light curve (scatter plot of an output image channel vs an input reference channel).

  Args:
    image (numpy.ndarray): The output image channel (luma, ...) as an array with shape (height, width).
    reference (numpy.ndarray): The input reference channel as an array with shape (height, width).
    maxpoints (int, optional): The maximum number of points in the scatter plot. The image and
      reference will be sampled accordingly (default 32768).
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    renderer (str, optional): The plotly renderer (default None = "jupyterlab").
  """
  if width <= 0: width = params.maxwidth
  # Prepare light curve.
  if image.ndim != 2 or reference.ndim != 2:
    raise ValueError("Error, image and reference must be 2D arrays.")
  if image.shape != reference.shape:
    raise ValueError("Error, image and reference must have the same shape.")
  input = reference.ravel()
  output = image.ravel()
  step = int(np.ceil(len(input)/maxpoints))
  # Set-up lines.
  mlinedot1 = dict(color = params.mlinecolor, dash = "dot", width = 1)
  mlinedashdot1 = dict(color = params.mlinecolor, dash = "dashdot", width = 1)
  # Plot light curve.
  figure = go.Figure()
  figure.add_trace(go.Scatter(x = input[::step], y = output[::step],
                              name = "Light curve", mode = "markers", marker = dict(size = 4, color = "lightslategray"), showlegend = False))
  figure.add_trace(go.Scatter(x = [0., 1.], y = [0., 0.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False))
  figure.add_trace(go.Scatter(x = [0., 1.], y = [1., 1.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False))
  figure.add_trace(go.Scatter(x = [0., 0.], y = [0., 1.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False))
  figure.add_trace(go.Scatter(x = [1., 1.], y = [0., 1.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False))
  figure.add_trace(go.Scatter(x = [0., 1.], y = [0., 1.], name = "", mode = "lines", line = mlinedot1, showlegend = False))
  figure.update_xaxes(title_text = "Input level", ticks = "inside", rangemode = "tozero")
  figure.update_yaxes(title_text = "Output level", ticks = "inside", rangemode = "tozero")
  layout = go.Layout(template = "plotly_dark",
                     width = width+params.lmargin+params.rmargin, height = width+params.bmargin+params.tmargin,
                     margin = go.layout.Margin(l = params.lmargin, r = params.rmargin, b = params.bmargin, t = params.tmargin, autoexpand = True))
  figure.update_layout(layout)
  figure.show(renderer)
